from .name2idx import C, V


class DifferentialEquation(object):
    """Kinetic equations"""

    def __init__(self, perturbation):
        self.perturbation = perturbation

    def diffeq(self, t, y, *x):
        dydt = [0] * V.NUM
        dydt[V.MP] = (
            +x[C.cvsP] * (y[V.BN] ** x[C.cn]) / (x[C.cKAP] ** x[C.cn] + y[V.BN] ** x[C.cn])
            - x[C.cvmP] * y[V.MP] / (x[C.cKmP] + y[V.MP])
            - x[C.ckdmp] * y[V.MP]
        )
        dydt[V.MC] = (
            +x[C.cvsC] * (y[V.BN] ** x[C.cn]) / (x[C.cKAC] ** x[C.cn] + y[V.BN] ** x[C.cn])
            - x[C.cvmC] * y[V.MC] / (x[C.cKmC] + y[V.MC])
            - x[C.ckdmc] * y[V.MC]
        )
        dydt[V.MB] = (
            +x[C.cvsB] * (x[C.cKIB] ** x[C.cm]) / (x[C.cKIB] ** x[C.cm] + y[V.BN] ** x[C.cm])
            - x[C.cvmB] * y[V.MB] / (x[C.cKmB] + y[V.MB])
            - x[C.ckdmb] * y[V.MB]
        )
        dydt[V.PC] = (
            +x[C.cksP] * y[V.MP]
            - x[C.cV1P] * y[V.PC] / (x[C.cKp] + y[V.PC])
            + x[C.cV2P] * y[V.PCP] / (x[C.cKdp] + y[V.PCP])
            + x[C.ck4] * y[V.PCC]
            - x[C.ck3] * y[V.PC] * y[V.CC]
            - x[C.ckdn] * y[V.PC]
        )
        dydt[V.CC] = (
            +x[C.cksC] * y[V.MC]
            - x[C.cV1C] * y[V.CC] / (x[C.cKp] + y[V.CC])
            + x[C.cV2C] * y[V.CCP] / (x[C.cKdp] + y[V.CCP])
            + x[C.ck4] * y[V.PCC]
            - x[C.ck3] * y[V.PC] * y[V.CC]
            - x[C.ckdnc] * y[V.CC]
        )
        dydt[V.PCP] = (
            +x[C.cV1P] * y[V.PC] / (x[C.cKp] + y[V.PC])
            - x[C.cV2P] * y[V.PCP] / (x[C.cKdp] + y[V.PCP])
            - x[C.cvdPC] * y[V.PCP] / (x[C.cKd] + y[V.PCP])
            - x[C.ckdn] * y[V.PCP]
        )
        dydt[V.CCP] = (
            +x[C.cV1C] * y[V.CC] / (x[C.cKp] + y[V.CC])
            - x[C.cV2C] * y[V.CCP] / (x[C.cKdp] + y[V.CCP])
            - x[C.cvdCC] * y[V.CCP] / (x[C.cKd] + y[V.CCP])
            - x[C.ckdn] * y[V.CCP]
        )
        dydt[V.PCC] = (
            -x[C.cV1PC] * y[V.PCC] / (x[C.cKp] + y[V.PCC])
            + x[C.cV2PC] * y[V.PCCP] / (x[C.cKdp] + y[V.PCCP])
            - x[C.ck4] * y[V.PCC]
            + x[C.ck3] * y[V.PC] * y[V.CC]
            + x[C.ck2] * y[V.PCN]
            - x[C.ck1] * y[V.PCC]
            - x[C.ckdn] * y[V.PCC]
        )
        dydt[V.PCN] = (
            -x[C.cV3PC] * y[V.PCN] / (x[C.cKp] + y[V.PCN])
            + x[C.cV4PC] * y[V.PCNP] / (x[C.cKdp] + y[V.PCNP])
            - x[C.ck2] * y[V.PCN]
            + x[C.ck1] * y[V.PCC]
            - x[C.ck7] * y[V.BN] * y[V.PCN]
            + x[C.ck8] * y[V.IN]
            - x[C.ckdn] * y[V.PCN]
        )
        dydt[V.PCCP] = (
            +x[C.cV1PC] * y[V.PCC] / (x[C.cKp] + y[V.PCC])
            - x[C.cV2PC] * y[V.PCCP] / (x[C.cKdp] + y[V.PCCP])
            - x[C.cvdPCC] * y[V.PCCP] / (x[C.cKd] + y[V.PCCP])
            - x[C.ckdn] * y[V.PCCP]
        )
        dydt[V.PCNP] = (
            +x[C.cV3PC] * y[V.PCN] / (x[C.cKp] + y[V.PCN])
            - x[C.cV4PC] * y[V.PCNP] / (x[C.cKdp] + y[V.PCNP])
            - x[C.cvdPCN] * y[V.PCNP] / (x[C.cKd] + y[V.PCNP])
            - x[C.ckdn] * y[V.PCNP]
        )
        dydt[V.BC] = (
            +x[C.cksB] * y[V.MB]
            - x[C.cV1B] * y[V.BC] / (x[C.cKp] + y[V.BC])
            + x[C.cV2B] * y[V.BCP] / (x[C.cKdp] + y[V.BCP])
            - x[C.ck5] * y[V.BC]
            + x[C.ck6] * y[V.BN]
            - x[C.ckdn] * y[V.BC]
        )
        dydt[V.BCP] = (
            +x[C.cV1B] * y[V.BC] / (x[C.cKp] + y[V.BC])
            - x[C.cV2B] * y[V.BCP] / (x[C.cKdp] + y[V.BCP])
            - x[C.cvdBC] * y[V.BCP] / (x[C.cKd] + y[V.BCP])
            - x[C.ckdn] * y[V.BCP]
        )
        dydt[V.BN] = (
            -x[C.cV3B] * y[V.BN] / (x[C.cKp] + y[V.BN])
            + x[C.cV4B] * y[V.BNP] / (x[C.cKdp] + y[V.BNP])
            + x[C.ck5] * y[V.BC]
            - x[C.ck6] * y[V.BN]
            - x[C.ck7] * y[V.BN] * y[V.PCN]
            + x[C.ck8] * y[V.IN]
            - x[C.ckdn] * y[V.BN]
        )
        dydt[V.BNP] = (
            +x[C.cV3B] * y[V.BN] / (x[C.cKp] + y[V.BN])
            - x[C.cV4B] * y[V.BNP] / (x[C.cKdp] + y[V.BNP])
            - x[C.cvdBN] * y[V.BNP] / (x[C.cKd] + y[V.BNP])
            - x[C.ckdn] * y[V.BNP]
        )
        dydt[V.IN] = (
            -x[C.ck8] * y[V.IN]
            + x[C.ck7] * y[V.BN] * y[V.PCN]
            - x[C.cvdIN] * y[V.IN] / (x[C.cKd] + y[V.IN])
            - x[C.ckdn] * y[V.IN]
        )

        return dydt


def param_values():
    # Parameter values
    x = [0] * C.NUM
    x[C.ck1] = 0.4
    x[C.ck2] = 0.2
    x[C.ck3] = 0.4
    x[C.ck4] = 0.2
    x[C.ck5] = 0.4
    x[C.ck6] = 0.2
    x[C.ck7] = 0.5
    x[C.ck8] = 0.1
    x[C.cKAP] = 0.7
    x[C.cKAC] = 0.6
    x[C.cKIB] = 2.2
    x[C.ckdmb] = 0.01
    x[C.ckdmc] = 0.01
    x[C.ckdmp] = 0.01
    x[C.ckdn] = 0.01
    x[C.ckdnc] = 0.12
    x[C.cKd] = 0.3
    x[C.cKdp] = 0.1
    x[C.cKp] = 0.1
    x[C.cKmB] = 0.4
    x[C.cKmC] = 0.4
    x[C.cKmP] = 0.31
    x[C.cksB] = 0.12
    x[C.cksC] = 1.6
    x[C.cksP] = 0.6
    x[C.cm] = 2.0
    x[C.cn] = 4.0
    x[C.cV1B] = 0.5
    x[C.cV1C] = 0.6
    x[C.cV1P] = 0.4
    x[C.cV1PC] = 0.4
    x[C.cV2B] = 0.1
    x[C.cV2C] = 0.1
    x[C.cV2P] = 0.3
    x[C.cV2PC] = 0.1
    x[C.cV3B] = 0.5
    x[C.cV3PC] = 0.4
    x[C.cV4B] = 0.2
    x[C.cV4PC] = 0.1
    x[C.cVphos] = 0.4
    x[C.cvdBC] = 0.5
    x[C.cvdBN] = 0.6
    x[C.cvdCC] = 0.7
    x[C.cvdIN] = 0.8
    x[C.cvdPC] = 0.7
    x[C.cvdPCC] = 0.7
    x[C.cvdPCN] = 0.7
    x[C.cvmB] = 0.8
    x[C.cvmC] = 1.0
    x[C.cvmP] = 1.1
    x[C.cvsB] = 1.0
    x[C.cvsC] = 1.1
    x[C.cvsP] = 1.5

    return x


def initial_values():
    # Value of the initial condition
    y0 = [0] * V.NUM

    y0[V.MB] = 9.0
    y0[V.BC] = 2.0
    y0[V.BN] = 1.9
    y0[V.MC] = 1.4
    y0[V.MP] = 1.6
    y0[V.PCN] = 1.0

    return y0


import csv
import os
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np

from .exec_model import ExecModel, ModelObject


@dataclass
class OptimizationResults(ExecModel):
    model: ModelObject

    def __post_init__(self) -> None:
        os.makedirs(
            os.path.join(
                self.model.path,
                "optimization_results",
            ),
            exist_ok=True,
        )

    def to_csv(self) -> None:
        """
        Save optimized parameters as CSV file format.

        Output
        ------
        optimization_results/optimized_params.csv
        optimization_results/optimized_initials.csv

        Examples
        --------
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> from biomass.result import OptimizationResults
        >>> model = Nakakuki_Cell_2010.create()
        >>> res = OptimizationResults(model)
        >>> res.to_csv()

        """
        n_file = self.get_executable()

        if len(self.model.sp.idx_params) > 0:
            optimized_params = np.empty(
                (len(self.model.sp.idx_params) + 2, len(n_file) + 1), dtype="<U21"
            )
            for i, parameter_index in enumerate(self.model.sp.idx_params):
                for j, nth_paramset in enumerate(sorted(n_file), start=1):
                    best_generation = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "generation.npy",
                        )
                    )
                    best_individual = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            f"fit_param{int(best_generation):d}.npy",
                        )
                    )
                    error = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "best_fitness.npy",
                        )
                    )
                    optimized_params[0, 0] = ""
                    optimized_params[1, 0] = "*Error*"
                    optimized_params[i + 2, 0] = self.model.parameters[parameter_index]
                    optimized_params[0, j] = str(nth_paramset)
                    optimized_params[1, j] = f"{error:8.3e}"
                    optimized_params[i + 2, j] = f"{best_individual[i]:8.3e}"
            with open(
                os.path.join(
                    self.model.path,
                    "optimization_results",
                    "optimized_params.csv",
                ),
                mode="w",
            ) as f:
                writer = csv.writer(f, lineterminator="\n")
                writer.writerows(optimized_params)
        if len(self.model.sp.idx_initials) > 0:
            optimized_initials = np.empty(
                (len(self.model.sp.idx_initials) + 2, len(n_file) + 1), dtype="<U21"
            )
            for i, species_index in enumerate(self.model.sp.idx_initials):
                for j, nth_paramset in enumerate(sorted(n_file), start=1):
                    best_generation = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "generation.npy",
                        )
                    )
                    best_individual = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            f"fit_param{int(best_generation):d}.npy",
                        )
                    )
                    error = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "best_fitness.npy",
                        )
                    )
                    optimized_initials[0, 0] = ""
                    optimized_initials[1, 0] = "*Error*"
                    optimized_initials[i + 2, 0] = self.model.species[species_index]
                    optimized_initials[0, j] = str(nth_paramset)
                    optimized_initials[1, j] = f"{error:8.3e}"
                    optimized_initials[
                        i + 2, j
                    ] = f"{best_individual[i+len(self.model.sp.idx_params)]:8.3e}"
            with open(
                os.path.join(
                    self.model.path,
                    "optimization_results",
                    "optimized_initials.csv",
                ),
                mode="w",
            ) as f:
                writer = csv.writer(f, lineterminator="\n")
                writer.writerows(optimized_initials)

    def dynamic_assessment(self, include_original: bool = False) -> None:
        """
        Compute objective values using estimated parameters.

        Parameters
        ----------
        include_original : bool (default: False)
            If True, an objective value simulated with original parameters
            will also be shown.

        Output
        ------
        fitness_assessment.csv

        Examples
        --------
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> from biomass.result import OptimizationResults
        >>> model = Nakakuki_Cell_2010.create()
        >>> res = OptimizationResults(model)
        >>> res.dynamic_assessment()

        """
        with open(
            os.path.join(
                self.model.path,
                "optimization_results",
                "fitness_assessment.csv",
            ),
            mode="w",
        ) as f:
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(["parameter set", "Objective value"])
            if include_original:
                x = self.model.pval()
                y0 = self.model.ival()
                obj_val = self.model.obj_func(None, x, y0)
                writer.writerow(["original", f"{obj_val:8.3e}"])
            n_file = self.get_executable()
            for paramset in sorted(n_file):
                optimized = self.load_param(paramset)
                obj_val = self.model.obj_func(None, *optimized)
                writer.writerow([f"{paramset:d}", f"{obj_val:8.3e}"])

    def trace_obj(self) -> None:
        """
        Visualize objective function traces for different optimization runs.

        Output
        ------
        obj_func_traces.pdf

        Examples
        --------
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> from biomass.result import OptimizationResults
        >>> model = Nakakuki_Cell_2010.create()
        >>> res = OptimizationResults(model)
        >>> res.trace_obj()

        """
        n_file = self.get_executable()
        # matplotlib
        plt.figure(figsize=(4, 3))
        plt.gca().spines["right"].set_visible(False)
        plt.gca().spines["top"].set_visible(False)
        plt.rcParams["font.size"] = 18
        plt.rcParams["lines.linewidth"] = 1.0
        # ---
        for paramset in n_file:
            with open(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{paramset:d}",
                    "optimization.log",
                ),
                mode="r",
            ) as f:
                traces = f.readlines()
            iters = []
            obj_val = []
            for line in traces:
                if line.startswith("Generation"):
                    iters.append(line.lstrip("Generation").split(":")[0])
                    obj_val.append(line.split("=")[-1].strip())
            plt.plot(
                [int(num) for num in iters],
                [float(val) for val in obj_val],
            )
        plt.xlabel("Iteration")
        plt.ylabel("Objective function value")
        plt.savefig(
            os.path.join(
                self.model.path,
                "optimization_results",
                "obj_func_traces.pdf",
            ),
            bbox_inches="tight",
        )



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation

observables = [
    "nuclear_IkBa",
    "nuclear_NFkB",
]


class NumericalSimulation(DifferentialEquation):
    """Simulate a model using scipy.integrate.solve_ivp

    Attributes
    ----------
    normalization : nested dict
        Keys for each observable
        ------------------------
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    """

    def __init__(self):
        super().__init__(perturbation={})
        self.normalization = {}

    t = range(200 + 1)

    # Experimental conditions
    conditions = ["TNFa", "TNFa_DCF"]

    simulations = np.empty((len(observables), len(t), len(conditions)))

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation
        for i, condition in enumerate(self.conditions):
            if condition == "TNFa":
                pass
            elif condition == "TNFa_DCF":
                x[C.uptake] = 1.0000
                x[C.TNF] = 1.0000
                x[C.trigger_iIkk] = 0.0195
                x[C.deact_TNFR] = 0.0010
                x[C.deact_ppIkk] = 0.1660
                x[C.deact_pnNfk] = 1000.0000
                x[C.act_Ikk_by_TNF] = 0.0347
                x[C.act_pIkk] = 0.1603
                x[C.act_Ikb_by_Ikk] = 0.1562
                x[C.act_Nfk_by_Ikk] = 0.6438
                x[C.act_Nfk_by_Ikk_complex] = 0.2816
                x[C.act_Ikb_complex] = 1.3897
                x[C.form_complex] = 2.8390
                x[C.form_complex_nuc] = 1000.0000
                x[C.ext_nNfkIkb] = 1000.0000
                x[C.Vnuc] = 1.0000
                x[C.split_NfkpIkb] = 0.0811
                x[C.split_NfkIkb] = 1.0000
                x[C.int_Nfk] = 0.0100
                x[C.int_Ikb] = 0.1226
                x[C.eta_int_pNfk] = 17.9585
                x[C.degrad_Ikb] = 0.6308
                x[C.degrad_mIkb] = 0.0053
                x[C.degrad_RnaA20] = 0.0089
                x[C.degrad_A20] = 0.0116
                x[C.prod_Ikb] = 1.0000
                x[C.prod_mIkb_by_nNfk] = 0.0020
                x[C.build_RnaA20] = 1.0000
                x[C.build_A20] = 0.0006
                x[C.shuttle_RnaA20] = 0.0119

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[observables.index("nuclear_IkBa"), :, i] = x[C.Vnuc] * (
                    sol.y[V.nNfkIkb, :] + sol.y[V.nIkb, :]
                )
                self.simulations[observables.index("nuclear_NFkB"), :, i] = x[C.Vnuc] * (
                    sol.y[V.pnNfk, :] + sol.y[V.nNfk, :] + sol.y[V.nNfkIkb, :]
                )


class ExperimentalData(object):
    """
    Set experimental data.

    Attributes
    ----------
    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        self.experiments = [None] * len(observables)
        self.error_bars = [None] * len(observables)

    def set_data(self):
        pass

    @staticmethod
    def get_timepoint(obs_name):
        if obs_name in observables:
            return []



import sys
from math import fabs, isnan, log
from typing import List, Optional, Union

import numpy as np
from scipy.integrate import simps


def _get_duration(
    timecourse: np.ndarray,
    below_threshold: float,
) -> int:
    """
    Calculation of the duration as the time it takes to decline below the threshold.

    Parameters
    ----------
    timecourse : array
        Simulated time course data.

    below_threshold : float (from 0.0 to 1.0)
        0.1 for 10% of its maximum.

    Returns
    -------
    duration : int

    """
    if not 0.0 < below_threshold < 1.0:
        raise ValueError("below_threshold must lie within (0.0, 1.0).")
    maximum_value = np.max(timecourse)
    t_max = np.argmax(timecourse)
    timecourse = timecourse - below_threshold * maximum_value
    timecourse[timecourse > 0.0] = -np.inf
    duration = np.argmax(timecourse[t_max:]) + t_max

    return duration


def get_signaling_metric(
    metric: str,
    timecourse: np.ndarray,
    options: dict,
) -> Optional[Union[int, float]]:
    """Quantification of cellular response.

    Parameters
    ----------
    metric : str
        Should be one of available_metrics.

    timecourse : array
        Simulated time course data.

    options: dict
        Options for detailed setting of signaling metric.

    Returns
    -------
    M* : int or float
        signaling_metric[i, j, k, l]

    """
    available_metrics = [
        "maximum",
        "minimum",
        "argmax",
        "argmin",
        "timepoint",
        "duration",
        "integral",
    ]
    if metric == "maximum":
        return np.max(timecourse)
    elif metric == "minimum":
        return np.min(timecourse)
    elif metric == "argmax":
        return np.argmax(timecourse)
    elif metric == "argmin":
        return np.argmin(timecourse)
    elif metric == "timepoint":
        return timecourse[options["timepoint"]]
    elif metric == "duration":
        return _get_duration(timecourse, options["duration"])
    elif metric == "integral":
        return simps(timecourse)
    else:
        raise ValueError(f"Available metrics are: {', '.join(available_metrics)}")


def dlnyi_dlnxj(
    signaling_metric: np.ndarray,
    n_file: List[int],
    perturbed_idx: List[int],
    observables: List[str],
    conditions: List[str],
    rate: float,
) -> np.ndarray:
    """
    Numerical computation of sensitivities using finite difference approximations
    with 1% changes in the reaction rates, parameter values or non-zero initial values.

    Parameters
    ----------
    signaling_metric : numpy array
        Signaling metric

    n_file : list of integers
        Optimized parameter sets in out/

    perturbed_idx : list of integers
        Indices of rate equations or non-zero initial values.

    observables : list of strings
        observables in observable.py

    conditions : list of strings
        Experimental conditions.

    rate : float ~ 1
        1.01 for 1% change.

    Returns
    -------
    sensitivity_coefficients: numpy array

    """
    sensitivity_coefficients = np.empty(
        (len(n_file), len(perturbed_idx), len(observables), len(conditions))
    )
    for i, _ in enumerate(n_file):
        for j, _ in enumerate(perturbed_idx):
            for k, _ in enumerate(observables):
                for l, _ in enumerate(conditions):
                    if isnan(signaling_metric[i, j, k, l]):
                        sensitivity_coefficients[i, j, k, l] = np.nan
                    elif (
                        fabs(signaling_metric[i, -1, k, l]) < sys.float_info.epsilon
                        or fabs(signaling_metric[i, j, k, l] - signaling_metric[i, -1, k, l])
                        < sys.float_info.epsilon
                        or (signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]) <= 0
                    ):
                        # 1. Signaling metric before adding perturbation is zero
                        # 2. Absolute change caused by perturbation is too small
                        # 3. Antilogarithm <= 0
                        sensitivity_coefficients[i, j, k, l] = 0.0
                    else:
                        sensitivity_coefficients[i, j, k, l] = log(
                            signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]
                        ) / log(rate)

    return sensitivity_coefficients



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation

observables = [
    "Ski",
    "Skil",
    "Dnmt3a",
    "Sox4",
    "Jun",
    "Smad7",
    "Klf10",
    "Bmp4",
    "Cxcl15",
    "Dusp5",
    "Tgfa",
    "Pdk4",
]


class NumericalSimulation(DifferentialEquation):
    """Simulate a model using scipy.integrate.ode
    Attributes
    ----------
    normalization : nested dict
        Keys for each observable
        ------------------------
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    """

    def __init__(self):
        super().__init__(perturbation={})
        self.normalization = {}

    t = range(600 + 1)  # min

    # Experimental conditions
    conditions = ["WT", "Smad2OE", "Smad3OE", "Smad4OE"]

    simulations = np.empty((len(observables), len(t), len(conditions)))

    @staticmethod
    def _set_gene_param(gene_name, x):
        if gene_name == "Ski":
            x[C.gene_turn] = x[C.Ski_turn]
            x[C.gene_act1] = x[C.Ski_act1]
            x[C.gene_act2] = x[C.Ski_act2]
            x[C.gene_act3] = x[C.Ski_act3]
            x[C.gene_inh1] = x[C.Ski_inh1]
            x[C.gene_inh2] = x[C.Ski_inh2]
            x[C.gene_inh3] = x[C.Ski_inh3]
        elif gene_name == "Skil":
            x[C.gene_turn] = x[C.Skil_turn]
            x[C.gene_act1] = x[C.Skil_act1]
            x[C.gene_act2] = x[C.Skil_act2]
            x[C.gene_act3] = x[C.Skil_act3]
            x[C.gene_inh1] = x[C.Skil_inh1]
            x[C.gene_inh2] = x[C.Skil_inh2]
            x[C.gene_inh3] = x[C.Skil_inh3]
        elif gene_name == "Dnmt3a":
            x[C.gene_turn] = x[C.Dnmt3a_turn]
            x[C.gene_act1] = x[C.Dnmt3a_act1]
            x[C.gene_act2] = x[C.Dnmt3a_act2]
            x[C.gene_act3] = x[C.Dnmt3a_act3]
            x[C.gene_inh1] = x[C.Dnmt3a_inh1]
            x[C.gene_inh2] = x[C.Dnmt3a_inh2]
            x[C.gene_inh3] = x[C.Dnmt3a_inh3]
        elif gene_name == "Sox4":
            x[C.gene_turn] = x[C.Sox4_turn]
            x[C.gene_act1] = x[C.Sox4_act1]
            x[C.gene_act2] = x[C.Sox4_act2]
            x[C.gene_act3] = x[C.Sox4_act3]
            x[C.gene_inh1] = x[C.Sox4_inh1]
            x[C.gene_inh2] = x[C.Sox4_inh2]
            x[C.gene_inh3] = x[C.Sox4_inh3]
        elif gene_name == "Jun":
            x[C.gene_turn] = x[C.Jun_turn]
            x[C.gene_act1] = x[C.Jun_act1]
            x[C.gene_act2] = x[C.Jun_act2]
            x[C.gene_act3] = x[C.Jun_act3]
            x[C.gene_inh1] = x[C.Jun_inh1]
            x[C.gene_inh2] = x[C.Jun_inh2]
            x[C.gene_inh3] = x[C.Jun_inh3]
        elif gene_name == "Smad7":
            x[C.gene_turn] = x[C.Smad7_turn]
            x[C.gene_act1] = x[C.Smad7_act1]
            x[C.gene_act2] = x[C.Smad7_act2]
            x[C.gene_act3] = x[C.Smad7_act3]
            x[C.gene_inh1] = x[C.Smad7_inh1]
            x[C.gene_inh2] = x[C.Smad7_inh2]
            x[C.gene_inh3] = x[C.Smad7_inh3]
        elif gene_name == "Klf10":
            x[C.gene_turn] = x[C.Klf10_turn]
            x[C.gene_act1] = x[C.Klf10_act1]
            x[C.gene_act2] = x[C.Klf10_act2]
            x[C.gene_act3] = x[C.Klf10_act3]
            x[C.gene_inh1] = x[C.Klf10_inh1]
            x[C.gene_inh2] = x[C.Klf10_inh2]
            x[C.gene_inh3] = x[C.Klf10_inh3]
        elif gene_name == "Bmp4":
            x[C.gene_turn] = x[C.Bmp4_turn]
            x[C.gene_act1] = x[C.Bmp4_act1]
            x[C.gene_act2] = x[C.Bmp4_act2]
            x[C.gene_act3] = x[C.Bmp4_act3]
            x[C.gene_inh1] = x[C.Bmp4_inh1]
            x[C.gene_inh2] = x[C.Bmp4_inh2]
            x[C.gene_inh3] = x[C.Bmp4_inh3]
        elif gene_name == "Cxcl15":
            x[C.gene_turn] = x[C.Cxcl15_turn]
            x[C.gene_act1] = x[C.Cxcl15_act1]
            x[C.gene_act2] = x[C.Cxcl15_act2]
            x[C.gene_act3] = x[C.Cxcl15_act3]
            x[C.gene_inh1] = x[C.Cxcl15_inh1]
            x[C.gene_inh2] = x[C.Cxcl15_inh2]
            x[C.gene_inh3] = x[C.Cxcl15_inh3]
        elif gene_name == "Dusp5":
            x[C.gene_turn] = x[C.Dusp5_turn]
            x[C.gene_act1] = x[C.Dusp5_act1]
            x[C.gene_act2] = x[C.Dusp5_act2]
            x[C.gene_act3] = x[C.Dusp5_act3]
            x[C.gene_inh1] = x[C.Dusp5_inh1]
            x[C.gene_inh2] = x[C.Dusp5_inh2]
            x[C.gene_inh3] = x[C.Dusp5_inh3]
        elif gene_name == "Tgfa":
            x[C.gene_turn] = x[C.Tgfa_turn]
            x[C.gene_act1] = x[C.Tgfa_act1]
            x[C.gene_act2] = x[C.Tgfa_act2]
            x[C.gene_act3] = x[C.Tgfa_act3]
            x[C.gene_inh1] = x[C.Tgfa_inh1]
            x[C.gene_inh2] = x[C.Tgfa_inh2]
            x[C.gene_inh3] = x[C.Tgfa_inh3]
        elif gene_name == "Pdk4":
            x[C.gene_turn] = x[C.Pdk4_turn]
            x[C.gene_act1] = x[C.Pdk4_act1]
            x[C.gene_act2] = x[C.Pdk4_act2]
            x[C.gene_act3] = x[C.Pdk4_act3]
            x[C.gene_inh1] = x[C.Pdk4_inh1]
            x[C.gene_inh2] = x[C.Pdk4_inh2]
            x[C.gene_inh3] = x[C.Pdk4_inh3]
        return x

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation

        for gene_name in observables:
            x = self._set_gene_param(gene_name, x)
            for i, condition in enumerate(self.conditions):
                if condition == "WT":
                    y0[V.S2] = x[C.S2tot]
                    y0[V.S3] = x[C.S3tot]
                    y0[V.S4] = x[C.S4tot]
                elif condition == "Smad2OE":
                    y0[V.S2] = 2 * x[C.S2tot]
                    y0[V.S3] = x[C.S3tot]
                    y0[V.S4] = x[C.S4tot]
                elif condition == "Smad3OE":
                    y0[V.S2] = x[C.S2tot]
                    y0[V.S3] = 16 * x[C.S3tot]
                    y0[V.S4] = x[C.S4tot]
                elif condition == "Smad4OE":
                    y0[V.S2] = x[C.S2tot]
                    y0[V.S3] = x[C.S3tot]
                    y0[V.S4] = 3 * x[C.S3tot]

                sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

                if sol is None:
                    return False
                else:
                    self.simulations[observables.index(gene_name), :, i] = np.log2(
                        sol.y[V.gene, :]
                    )


class ExperimentalData(object):
    """
    Set experimental data.
    Attributes
    ----------
    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        self.experiments = [None] * len(observables)
        self.error_bars = [None] * len(observables)

    def set_data(self):
        pass

    @staticmethod
    def get_timepoint(obs_name):
        if obs_name in observables:
            return []



import numpy as np

from biomass.dynamics.solver import get_steady_state, solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation

observables = [
    "Phosphorylated_MEKc",
    "Phosphorylated_ERKc",
    "Phosphorylated_RSKw",
    "Phosphorylated_CREBw",
    "dusp_mRNA",
    "cfos_mRNA",
    "cFos_Protein",
    "Phosphorylated_cFos",
]


class NumericalSimulation(DifferentialEquation):
    """Simulate a model using scipy.integrate.solve_ivp

    Attributes
    ----------
    normalization : nested dict
        Keys for each observable
        ------------------------
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    """

    def __init__(self):
        super().__init__(perturbation={})
        self.normalization = {}
        for observable in observables:
            self.normalization[observable] = {"timepoint": None, "condition": []}

    t = range(5401)  # 0, 1, 2, ..., 5400 (Unit: sec.)

    # Experimental conditions
    conditions = ["EGF", "HRG"]

    simulations = np.empty((len(observables), len(t), len(conditions)))

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation
        # get steady state
        x[C.Ligand] = x[C.no_ligand]  # No ligand
        y0 = get_steady_state(self.diffeq, y0, tuple(x))
        if not y0:
            return False
        # add ligand
        for i, condition in enumerate(self.conditions):
            if condition == "EGF":
                x[C.Ligand] = x[C.EGF]
            elif condition == "HRG":
                x[C.Ligand] = x[C.HRG]

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[observables.index("Phosphorylated_MEKc"), :, i] = sol.y[
                    V.ppMEKc, :
                ]
                self.simulations[observables.index("Phosphorylated_ERKc"), :, i] = (
                    sol.y[V.pERKc, :] + sol.y[V.ppERKc, :]
                )
                self.simulations[observables.index("Phosphorylated_RSKw"), :, i] = sol.y[
                    V.pRSKc, :
                ] + sol.y[V.pRSKn, :] * (x[C.Vn] / x[C.Vc])
                self.simulations[observables.index("Phosphorylated_CREBw"), :, i] = sol.y[
                    V.pCREBn, :
                ] * (x[C.Vn] / x[C.Vc])
                self.simulations[observables.index("dusp_mRNA"), :, i] = sol.y[V.duspmRNAc, :]
                self.simulations[observables.index("cfos_mRNA"), :, i] = sol.y[V.cfosmRNAc, :]
                self.simulations[observables.index("cFos_Protein"), :, i] = (
                    (sol.y[V.pcFOSn, :] + sol.y[V.cFOSn, :]) * (x[C.Vn] / x[C.Vc])
                    + sol.y[V.cFOSc, :]
                    + sol.y[V.pcFOSc, :]
                )
                self.simulations[observables.index("Phosphorylated_cFos"), :, i] = (
                    sol.y[V.pcFOSn, :] * (x[C.Vn] / x[C.Vc]) + sol.y[V.pcFOSc, :]
                )


class ExperimentalData(object):
    """
    Set experimental data.

    Attributes
    ----------
    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        self.experiments = [None] * len(observables)
        self.error_bars = [None] * len(observables)

    def set_data(self):
        self.experiments[observables.index("Phosphorylated_MEKc")] = {
            "EGF": [0.000, 0.773, 0.439, 0.252, 0.130, 0.087, 0.080, 0.066],
            "HRG": [0.000, 0.865, 1.000, 0.837, 0.884, 0.920, 0.875, 0.789],
        }
        self.error_bars[observables.index("Phosphorylated_MEKc")] = {
            "EGF": [
                sd / np.sqrt(3) for sd in [0.000, 0.030, 0.048, 0.009, 0.009, 0.017, 0.012, 0.008]
            ],
            "HRG": [
                sd / np.sqrt(3) for sd in [0.000, 0.041, 0.000, 0.051, 0.058, 0.097, 0.157, 0.136]
            ],
        }

        self.experiments[observables.index("Phosphorylated_ERKc")] = {
            "EGF": [0.000, 0.867, 0.799, 0.494, 0.313, 0.266, 0.200, 0.194],
            "HRG": [0.000, 0.848, 1.000, 0.971, 0.950, 0.812, 0.747, 0.595],
        }
        self.error_bars[observables.index("Phosphorylated_ERKc")] = {
            "EGF": [
                sd / np.sqrt(3) for sd in [0.000, 0.137, 0.188, 0.126, 0.096, 0.087, 0.056, 0.012]
            ],
            "HRG": [
                sd / np.sqrt(3) for sd in [0.000, 0.120, 0.000, 0.037, 0.088, 0.019, 0.093, 0.075]
            ],
        }

        self.experiments[observables.index("Phosphorylated_RSKw")] = {
            "EGF": [0, 0.814, 0.812, 0.450, 0.151, 0.059, 0.038, 0.030],
            "HRG": [0, 0.953, 1.000, 0.844, 0.935, 0.868, 0.779, 0.558],
        }
        self.error_bars[observables.index("Phosphorylated_RSKw")] = {
            "EGF": [
                sd / np.sqrt(3) for sd in [0, 0.064, 0.194, 0.030, 0.027, 0.031, 0.043, 0.051]
            ],
            "HRG": [
                sd / np.sqrt(3) for sd in [0, 0.230, 0.118, 0.058, 0.041, 0.076, 0.090, 0.077]
            ],
        }

        self.experiments[observables.index("Phosphorylated_cFos")] = {
            "EGF": [0, 0.060, 0.109, 0.083, 0.068, 0.049, 0.027, 0.017],
            "HRG": [0, 0.145, 0.177, 0.158, 0.598, 1.000, 0.852, 0.431],
        }
        self.error_bars[observables.index("Phosphorylated_cFos")] = {
            "EGF": [
                sd / np.sqrt(3) for sd in [0, 0.003, 0.021, 0.013, 0.016, 0.007, 0.003, 0.002]
            ],
            "HRG": [
                sd / np.sqrt(3) for sd in [0, 0.010, 0.013, 0.001, 0.014, 0.000, 0.077, 0.047]
            ],
        }

        # ----------------------------------------------------------------------

        self.experiments[observables.index("Phosphorylated_CREBw")] = {
            "EGF": [0, 0.446, 0.030, 0.000, 0.000],
            "HRG": [0, 1.000, 0.668, 0.460, 0.340],
        }
        self.error_bars[observables.index("Phosphorylated_CREBw")] = {
            "EGF": [sd / np.sqrt(3) for sd in [0, 0.0, 0.0, 0.0, 0.0]],
            "HRG": [sd / np.sqrt(3) for sd in [0, 0.0, 0.0, 0.0, 0.0]],
        }
        # ----------------------------------------------------------------------

        self.experiments[observables.index("cfos_mRNA")] = {
            "EGF": [0, 0.181, 0.476, 0.518, 0.174, 0.026, 0.000],
            "HRG": [0, 0.353, 0.861, 1.000, 0.637, 0.300, 0.059],
        }
        self.error_bars[observables.index("cfos_mRNA")] = {
            "EGF": [sd / np.sqrt(3) for sd in [0.017, 0.004, 0.044, 0.004, 0.023, 0.007, 0.008]],
            "HRG": [sd / np.sqrt(3) for sd in [0.017, 0.006, 0.065, 0.044, 0.087, 0.023, 0.001]],
        }
        # ----------------------------------------------------------------------

        self.experiments[observables.index("cFos_Protein")] = {
            "EGF": [0, 0.078, 0.216, 0.240, 0.320, 0.235],
            "HRG": [0, 0.089, 0.552, 0.861, 1.000, 0.698],
        }
        self.error_bars[observables.index("cFos_Protein")] = {
            "EGF": [sd / np.sqrt(3) for sd in [0, 0.036, 0.028, 0.056, 0.071, 0.048]],
            "HRG": [sd / np.sqrt(3) for sd in [0, 0.021, 0.042, 0.063, 0.000, 0.047]],
        }

        self.experiments[observables.index("dusp_mRNA")] = {
            "EGF": [0.000, 0.177, 0.331, 0.214, 0.177, 0.231],
            "HRG": [0.000, 0.221, 0.750, 1.000, 0.960, 0.934],
        }
        self.error_bars[observables.index("dusp_mRNA")] = {
            "EGF": [sd / np.sqrt(3) for sd in [0.033, 0.060, 0.061, 0.032, 0.068, 0.050]],
            "HRG": [sd / np.sqrt(3) for sd in [0.027, 0.059, 0.094, 0.124, 0.113, 0.108]],
        }

    @staticmethod
    def get_timepoint(obs_name):
        if obs_name in [
            "Phosphorylated_MEKc",
            "Phosphorylated_ERKc",
            "Phosphorylated_RSKw",
            "Phosphorylated_cFos",
        ]:
            return [0, 300, 600, 900, 1800, 2700, 3600, 5400]  # (Unit: sec.)

        elif obs_name == "Phosphorylated_CREBw":
            return [0, 600, 1800, 3600, 5400]

        elif obs_name == "cfos_mRNA":
            return [0, 600, 1200, 1800, 2700, 3600, 5400]

        elif obs_name in ["cFos_Protein", "dusp_mRNA"]:
            return [0, 900, 1800, 2700, 3600, 5400]



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "ck1",
    "ck2",
    "ck3",
    "ck4",
    "ck5",
    "ck6",
    "ck7",
    "ck8",
    "cKAP",
    "cKAC",
    "cKIB",
    "ckdmb",
    "ckdmc",
    "ckdmp",
    "ckdn",
    "ckdnc",
    "cKd",
    "cKdp",
    "cKp",
    "cKmB",
    "cKmC",
    "cKmP",
    "cksB",
    "cksC",
    "cksP",
    "cm",
    "cn",
    "cV1B",
    "cV1C",
    "cV1P",
    "cV1PC",
    "cV2B",
    "cV2C",
    "cV2P",
    "cV2PC",
    "cV3B",
    "cV3PC",
    "cV4B",
    "cV4PC",
    "cVphos",
    "cvdBC",
    "cvdBN",
    "cvdCC",
    "cvdIN",
    "cvdPC",
    "cvdPCC",
    "cvdPCN",
    "cvmB",
    "cvmC",
    "cvmP",
    "cvsB",
    "cvsC",
    "cvsP",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



from .name2idx import C, V


class DifferentialEquation(object):
    def __init__(self, perturbation):
        self.perturbation = perturbation

    def diffeq(self, t, y, *x):
        """Table S6: List of differential equations of the reduced model"""
        dydt = [0] * V.NUM

        dydt[V.TNFR] = 1 * x[C.uptake] * x[C.TNF] - 1 * x[C.deact_TNFR] * y[V.TNFR]
        dydt[V.Ikk] = -1 * (x[C.act_Ikk_by_TNF] * y[V.TNFR] * y[V.Ikk]) + 1 * (x[C.trigger_iIkk] * y[V.iIkk])
        dydt[V.pIkk] = 1 * (x[C.act_Ikk_by_TNF] * y[V.TNFR] * y[V.Ikk]) - 1 * (x[C.act_pIkk] * y[V.pIkk])
        dydt[V.ppIkk] = 1 * (x[C.act_pIkk] * y[V.pIkk]) - 1 * (x[C.deact_ppIkk] * y[V.ppIkk])
        dydt[V.iIkk] = 1 * (x[C.deact_ppIkk] * y[V.ppIkk]) - 1 * (x[C.trigger_iIkk] * y[V.iIkk])
        dydt[V.NfkIkb] = (
            -1 * ((x[C.act_Ikb_by_Ikk]) * y[V.pIkk] * y[V.NfkIkb])
            - 1 * ((x[C.act_Nfk_by_Ikk]) * y[V.pIkk] * y[V.NfkIkb])
            + 1 * (x[C.form_complex] * y[V.Nfk] * y[V.Ikb])
            + 1 * (x[C.ext_nNfkIkb] * y[V.nNfkIkb]) * (x[C.Vnuc] / 1)
        )
        dydt[V.NfkpIkb] = (
            1 * ((x[C.act_Ikb_by_Ikk]) * y[V.pIkk] * y[V.NfkIkb])
            - 1 * ((x[C.act_Nfk_by_Ikk_complex]) * y[V.pIkk] * y[V.NfkpIkb])
            - 1 * (x[C.split_NfkpIkb] * y[V.NfkpIkb])
        )
        dydt[V.pNfkIkb] = (
            -1 * ((x[C.act_Ikb_complex]) * y[V.pNfkIkb])
            - 1 * ((x[C.act_Ikb_by_Ikk]) * y[V.pIkk] * y[V.pNfkIkb])
            + 1 * ((x[C.act_Nfk_by_Ikk]) * y[V.pIkk] * y[V.NfkIkb])
        )
        dydt[V.pNfkpIkb] = (
            1 * ((x[C.act_Ikb_complex]) * y[V.pNfkIkb])
            + 1 * ((x[C.act_Ikb_by_Ikk]) * y[V.pIkk] * y[V.pNfkIkb])
            + 1 * ((x[C.act_Nfk_by_Ikk_complex]) * y[V.pIkk] * y[V.NfkpIkb])
            - 1 * (x[C.split_NfkIkb] * y[V.pNfkpIkb])
        )
        dydt[V.pNfk] = 1 * (x[C.split_NfkIkb] * y[V.pNfkpIkb]) - 1 * ((x[C.int_Nfk] * x[C.eta_int_pNfk]) * y[V.pNfk])
        dydt[V.Nfk] = (
            1 * (x[C.split_NfkpIkb] * y[V.NfkpIkb])
            - 1 * (x[C.form_complex] * y[V.Nfk] * y[V.Ikb])
            - 1 * ((x[C.int_Nfk]) * y[V.Nfk])
        )
        dydt[V.pIkb] = (
            1 * (x[C.split_NfkpIkb] * y[V.NfkpIkb])
            + 1 * (x[C.split_NfkIkb] * y[V.pNfkpIkb])
            - 1 * (x[C.degrad_Ikb] * y[V.pIkb])
        )
        dydt[V.Ikb] = (
            -1 * (x[C.form_complex] * y[V.Nfk] * y[V.Ikb])
            + 1 * (x[C.prod_Ikb] * y[V.mIkb])
            - 1 * (x[C.int_Ikb] * y[V.Ikb])
        )
        dydt[V.mIkb] = 1 * (x[C.prod_mIkb_by_nNfk] * y[V.nNfk]) - 1 * (x[C.degrad_mIkb] * y[V.mIkb])
        dydt[V.nIkb] = 1 * (x[C.int_Ikb] * y[V.Ikb]) * (1 / x[C.Vnuc]) - 1 * (
            x[C.form_complex_nuc] * y[V.nNfk] * y[V.nIkb]
        )
        dydt[V.pnNfk] = 1 * ((x[C.int_Nfk] * x[C.eta_int_pNfk]) * y[V.pNfk]) * (1 / x[C.Vnuc]) - 1 * (
            x[C.deact_pnNfk] * y[V.pnNfk]
        )
        dydt[V.nNfk] = (
            1 * ((x[C.int_Nfk]) * y[V.Nfk]) * (1 / x[C.Vnuc])
            + 1 * (x[C.deact_pnNfk] * y[V.pnNfk])
            - 1 * (x[C.form_complex_nuc] * y[V.nNfk] * y[V.nIkb])
        )
        dydt[V.nNfkIkb] = 1 * (x[C.form_complex_nuc] * y[V.nNfk] * y[V.nIkb]) - 1 * (x[C.ext_nNfkIkb] * y[V.nNfkIkb])
        dydt[V.RnaA20_1] = 1 * (x[C.build_RnaA20] * y[V.nNfk]) - 1 * (x[C.shuttle_RnaA20] * y[V.RnaA20_1])
        dydt[V.RnaA20] = 1 * (x[C.shuttle_RnaA20] * y[V.RnaA20_1]) - 1 * (x[C.degrad_RnaA20] * y[V.RnaA20])
        dydt[V.A20] = 1 * (x[C.build_A20] * y[V.RnaA20]) - 1 * (x[C.degrad_A20] * y[V.A20])

        return dydt


def param_values():
    x = [0] * C.NUM

    x[C.uptake] = 1.0000
    x[C.TNF] = 1.0000
    x[C.trigger_iIkk] = 0.0041
    x[C.deact_TNFR] = 0.0010
    x[C.deact_ppIkk] = 0.1660
    x[C.deact_pnNfk] = 1000.0000
    x[C.act_Ikk_by_TNF] = 0.0714
    x[C.act_pIkk] = 0.0648
    x[C.act_Ikb_by_Ikk] = 0.3980
    x[C.act_Nfk_by_Ikk] = 0.6438
    x[C.act_Nfk_by_Ikk_complex] = 0.2816
    x[C.act_Ikb_complex] = 1.3897
    x[C.form_complex] = 2.8390
    x[C.form_complex_nuc] = 1000.0000
    x[C.ext_nNfkIkb] = 1000.0000
    x[C.Vnuc] = 1.0000
    x[C.split_NfkpIkb] = 0.0811
    x[C.split_NfkIkb] = 1.0000
    x[C.int_Nfk] = 0.0100
    x[C.int_Ikb] = 0.1226
    x[C.eta_int_pNfk] = 17.9585
    x[C.degrad_Ikb] = 0.6308
    x[C.degrad_mIkb] = 0.0313
    x[C.degrad_RnaA20] = 0.0089
    x[C.degrad_A20] = 0.0116
    x[C.prod_Ikb] = 1.0000
    x[C.prod_mIkb_by_nNfk] = 0.0047
    x[C.build_RnaA20] = 1.0000
    x[C.build_A20] = 0.0006
    x[C.shuttle_RnaA20] = 0.0311

    return x


def initial_values():
    y0 = [0] * V.NUM

    y0[V.Ikk] = 1.0
    y0[V.NfkIkb] = 1.0

    return y0


"""BioMASS core functions"""
import multiprocessing
import os
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from typing import Any, Optional

from .analysis import InitialConditionSensitivity, ParameterSensitivity, ReactionSensitivity
from .dynamics import SignalingSystems
from .estimation import GeneticAlgorithmContinue, GeneticAlgorithmInit
from .exec_model import ModelObject

__all__ = ["Model", "optimize", "optimize_continue", "run_simulation", "run_analysis"]


@dataclass
class Model(object):
    """
    The BioMASS model object.

    Attributes
    ----------
    pkg_name: str
        Path (dot-sepalated) to a biomass model directory.
        Use '__package__'.
    """

    pkg_name: str

    def _load_model(self) -> Any:
        try:
            biomass_model = import_module(self.pkg_name)
            return biomass_model
        except ImportError:
            p = Path(self.pkg_name.replace(".", os.sep))
            print(f"cannot import '{p.name}' from '{p.parent}'.")

    def create(self, show_info: bool = False) -> ModelObject:
        """
        Build a biomass model.

        Parameters
        ----------
        show_info : bool (default: False)
            Set to 'True' to print the information related to model size.

        Examples
        --------
        >>> from biomass import Model
        >>> import your_model
        >>> model = Model(your_model.__package__).create()
        """
        model = ModelObject(self.pkg_name.replace(".", os.sep), self._load_model())
        if model.sim.normalization:
            for obs_name in model.obs:
                if (
                    isinstance(model.sim.normalization[obs_name]["timepoint"], int)
                    and not model.sim.t[0]
                    <= model.sim.normalization[obs_name]["timepoint"]
                    <= model.sim.t[-1]
                ):
                    raise ValueError("Normalization timepoint must lie within sim.t.")
                if not model.sim.normalization[obs_name]["condition"]:
                    model.sim.normalization[obs_name]["condition"] = model.sim.conditions
                else:
                    for c in model.sim.normalization[obs_name]["condition"]:
                        if c not in model.sim.conditions:
                            raise ValueError(
                                f"Normalization condition '{c}' is not defined in sim.conditions."
                            )
        if show_info:
            model_name = Path(model.path).name
            print(
                f"{model_name} information\n" + ("-" * len(model_name)) + "------------\n"
                f"{len(model.species):d} species\n"
                f"{len(model.parameters):d} parameters, "
                f"of which {len(model.sp.idx_params):d} to be estimated"
            )
        return model


def _check_optional_arguments(
    end: Optional[int],
    options: Optional[dict],
) -> None:
    if options is None:
        pass
    elif isinstance(options, dict):
        if options["local_search_method"].lower() not in ["mutation", "powell", "de"]:
            raise ValueError(
                f"'{options['local_search_method']}': "
                "Invalid local_search_method. Should be one of ['mutation', 'Powell', 'DE']"
            )
        elif (
            isinstance(end, int)
            and options["local_search_method"].lower() == "de"
            and options["workers"] != 1
        ):
            raise AssertionError(
                "daemonic processes are not allowed to have children. Set options['workers'] to 1."
            )
    else:
        raise TypeError("options must be dict or None.")


def optimize(
    model: ModelObject,
    start: int,
    end: Optional[int] = None,
    options: Optional[dict] = None,
) -> None:
    """
    Run GA for parameter estimation.

    Paremters
    ---------
    model : ModelObject
        Model for parameter estimation.

    start : int
        Index of parameter set to estimate.

    end : int, optional
        When `end` is specified, parameter sets from `start` to `end` will be estimated.

    options : dict, optional
        popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has popsize * len(search_param) individuals.

        max_generation : int (default: 10000)
            Stop optimization if Generation > max_generation.

        initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).

        allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= allowable_error.

        local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of
            * 'mutation' : NDM/MGG
            * 'Powell' : Modified Powell method
            * 'DE' : Differential Evolution (strategy: best2bin)

        n_children : int (default: 200)
            (method='mutation') The number of children generated in NDM/MGG.

        maxiter : int (default: 10)
            (method='Powell' or 'DE') The maximum number of iterations
            over which the entire population is evolved.

        workers : int (default: -1 if `end` is None else 1)
            (method='DE') The population is subdivided into workers sections and
            evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.

        overwrite : bool (default: False)
            If True, the out/n folder will be overwritten.

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import optimize
    >>> model = Nakakuki_Cell_2010.create()
    >>> optimize(
            model=model, start=1, end=10,
            options={
                'max_generation': 10000,
                'allowable_error': 0.5
            }
        )

    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 10000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 200)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if end is None else 1)
    options.setdefault("overwrite", False)

    _check_optional_arguments(end, options)

    ga_init = GeneticAlgorithmInit(model, **options)
    if end is None:
        ga_init.run(int(start))
    else:
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(
                ga_init.run,
                range(int(start), int(end) + 1),
            ):
                pass


def optimize_continue(
    model: ModelObject,
    start: int,
    end: Optional[int] = None,
    options: Optional[dict] = None,
) -> None:
    """
    Continue running GA from where you stopped in the last parameter search.

    Paremters
    ---------
    model : ModelObject
        Model for parameter estimation.

    start : int
        Index of parameter set to estimate.

    end : int, optional
        When `end` is specified, parameter sets from `start` to `end` will be estimated.

    options : dict, optional
        popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has popsize * len(search_param) individuals.

        max_generation : int (default: 15000)
            Stop optimization if Generation > max_generation.

        initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).

        allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= allowable_error.

        local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of
            * 'mutation' : NDM/MGG
            * 'Powell' : Modified Powell method
            * 'DE' : Differential Evolution (strategy: best2bin)

        n_children : int (default: 200)
            (method='mutation') The number of children generated in NDM/MGG.

        maxiter : int (default: 10)
            (method='Powell' or 'DE') The maximum number of iterations
            over which the entire population is evolved.

        workers : int (default: -1 if `end` is None else 1)
            (method='DE') The population is subdivided into workers sections and
            evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.

        p0_bounds : list of floats (default: [0.1, 10.0])
            Generate initial population using best parameter values in the last
            parameter search.
                - lower_bound = po_bounds[0] * best_parameter_value
                - upper_bound = p0_bounds[1] * best_parameter_value

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import optimize_continue
    >>> model = Nakakuki_Cell_2010.create()
    >>> optimize_continue(
            model=model, start=1, end=10,
            options={
                'max_generation': 20000,
                'allowable_error': 0.5
            }
        )

    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 15000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 200)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if end is None else 1)
    options.setdefault("p0_bounds", [0.1, 10.0])

    _check_optional_arguments(end, options)

    ga_continue = GeneticAlgorithmContinue(model, **options)
    if end is None:
        ga_continue.run(int(start))
    else:
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(
                ga_continue.run,
                range(int(start), int(end) + 1),
            ):
                pass


def run_simulation(
    model: ModelObject,
    viz_type: str,
    show_all: bool = False,
    stdev: bool = False,
    save_format: str = "pdf",
    param_range: Optional[dict] = None,
) -> None:
    """
    Simulate ODE model with estimated parameter values.

    Parameters
    ----------
    model : ModelObject
        Model for simulation.

    viz_type : str
        * 'average':
            The average of simulation results with parameter sets in "out/".
        * 'best':
            The best simulation result in "out/", simulation with
            "best_fit_param".
        * 'original':
            Simulation with the default parameters and initial values
            defined in "set_model.py".
        * 'n(=1,2,...)':
            Use the parameter set in "out/n/".
        * 'experiment'
            Draw the experimental data written in observable.py without
            simulation results.

    show_all : bool
        Whether to show all simulation results.

    stdev : bool
        If True, the standard deviation of simulated values will be shown
        (only available for 'average' visualization type).

    save_format : str (default: "pdf")
        Either "png" or "pdf", indicating whether to save figures
        as png or pdf format.

    param_range : dict, optional
        orientation : str (default: 'portrait')
            Either 'portrait' or 'landscape'.

        distribution : str (default: 'boxenplot')
            Either 'boxplot' or 'boxenplot'.

        scatter : bool (default: False)
            If True, draw a stripplot.

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import run_simulation
    >>> model = Nakakuki_Cell_2010.create()
    >>> run_simulation(
            model,
            viz_type='average',
            show_all=False,
            stdev=True,
            save_format="png",
        )

    """
    if viz_type not in ["best", "average", "original", "experiment"] and not viz_type.isdecimal():
        raise ValueError(
            "Available viz_type are: 'best','average','original','experiment','n(=1, 2, ...)'"
        )

    if save_format not in ["pdf", "png"]:
        raise ValueError("save_format must be either 'pdf' or 'png'.")

    if param_range is None:
        param_range = {}
    param_range.setdefault("orientation", "portrait")
    param_range.setdefault("distribution", "boxenplot")
    param_range.setdefault("scatter", False)

    if param_range["orientation"] not in ["portrait", "landscape"]:
        raise ValueError("Available param_range['orientation'] are: 'portrait' or 'landscape'.")
    if param_range["distribution"] not in ["boxplot", "boxenplot"]:
        raise ValueError("Available param_range['distribution'] are: 'boxplot' or 'boxenplot'.")
    if not isinstance(param_range["scatter"], bool):
        raise TypeError("param_range['scatter'] must be a boolean.")

    SignalingSystems(model).simulate_all(
        viz_type=viz_type,
        show_all=show_all,
        stdev=stdev,
        save_format=save_format,
        param_range=param_range,
    )


def run_analysis(
    model: ModelObject,
    target: str,
    metric: str = "integral",
    style: str = "barplot",
    save_format: str = "pdf",
    options: Optional[dict] = None,
) -> None:
    """
    Perform sensitivity analysis to identify critical parameters, species or
    reactions in the complex biological network.

    The sensitivity S(y,x) was calculated according to the following equation:
    S(y,x) = d ln(yi) / d ln (xj), where yi is the signaling metric and xj is
    each nonzero species, parameter value or reaction rate.

    Paremters
    ---------
    model : ModelObject
        Model for sensitivity analysis.

    target : str
        * 'reaction'
        * 'initial_condition'
        * 'parameter'

    metric : str (default: 'integral')
        * 'maximum' : The maximum value.
        * 'minimum' : The minimum value.
        * 'argmax' : The time to reach the maximum value.
        * 'argmin' : The time to reach the minimum value.
        * 'timepoint' : The simulated value at the time point set via options['timepoint'].
        * 'duration' :  The time it takes to decline below the threshold set via options['duration'].
        * 'integral' : The integral of concentration over the observation time.

    style : str (default: 'barplot')
        * 'barplot'
        * 'heatmap'

    save_format : str (default: "pdf")
        Either "png" or "pdf", indicating whether to save figures
        as png or pdf format.

    options : dict, optional
        show_indices : bool (default: True)
            (target == 'reaction') Set to True to put reaction index on each bar.

        excluded_params : list of strings
            (target == 'parameter') List of parameters which are not used for analysis.

        excluded_initials : list of strings
            (target == 'initial_condition') List of species which are not used for analysis.

        timepoint : int (default: model.sim.t[-1])
            (metric=='timepoint') Which timepoint to use.

        duration : float (default: 0.5)
            (metric=='duration') 0.1 for 10% of its maximum.

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import run_analysis
    >>> model = Nakakuki_Cell_2010.create()

    1. Parameter
    >>> run_analysis(
            model,
            target='parameter',
            options = {
                'excluded_params': [
                    'a', 'Vn', 'Vc', 'Ligand', 'EGF', 'HRG', 'no_ligand'
                ]
            }
        )

    2. Initial condition
    >>> run_analysis(
            model,
            target='initial_condition',
        )

    3. Reaction
    >>> run_analysis(
            model,
            target='reaction',
        )

    """
    if save_format not in ["pdf", "png"]:
        raise ValueError("save_format must be either 'pdf' or 'png'.")

    if options is None:
        options = {}
    options.setdefault("show_indices", True)
    options.setdefault("excluded_params", [])
    options.setdefault("excluded_initials", [])
    options.setdefault("timepoint", model.sim.t[-1])
    options.setdefault("duration", 0.5)

    if not model.sim.t[0] <= options["timepoint"] <= model.sim.t[-1]:
        raise ValueError("options['timepooint'] must lie within sim.t.")
    if not 0.0 < options["duration"] < 1.0:
        raise ValueError("options['duration'] must lie within (0, 1).")

    if target == "reaction":
        ReactionSensitivity(model).analyze(
            metric=metric,
            style=style,
            save_format=save_format,
            options=options,
        )
    elif target == "parameter":
        ParameterSensitivity(model).analyze(
            metric=metric,
            style=style,
            save_format=save_format,
            options=options,
        )
    elif target == "initial_condition":
        InitialConditionSensitivity(model).analyze(
            metric=metric,
            style=style,
            save_format=save_format,
            options=options,
        )
    else:
        raise ValueError(
            "Available targets are: '{}".format(
                "', '".join(["reaction", "parameter", "initial_condition"]) + "'."
            )
        )



from typing import List, NoReturn

import numpy as np


def initialize_search_param(
    parameters: List[str],
    species: List[str],
    param_values: list,
    initial_values: list,
    estimated_params: List[int],
    estimated_initials: List[int],
) -> np.ndarray:
    """
    Initialize search_param.

    Parameters
    ----------
    parameters : List[string]
        Names of model parameters.

    species : List[string]
        Names of model species.

    param_values : List[float]
        Parameter values.

    initial_values : List[float]
        Initial values.

    estimated_params : List[int]
        Indices of parameters to be estimated.

    estimated_initials : List[int]
        Indices of initial conditions to be estimated.

    Returns
    -------
    search_param : numpy ndarray
        Numerical values of parameters and/or initial conditions to be estimated.
    """
    if len(estimated_params) != len(set(estimated_params)):
        raise ValueError(
            "Duplicate parameters (C.): {}".format(
                ", ".join(
                    [
                        parameters[idx]
                        for idx in [
                            name
                            for name in set(estimated_params)
                            if estimated_params.count(name) > 1
                        ]
                    ]
                )
            )
        )

    if len(estimated_initials) != len(set(estimated_initials)):
        raise ValueError(
            "Duplicate initial conditions (V.): {}".format(
                ", ".join(
                    [
                        species[idx]
                        for idx in [
                            name
                            for name in set(estimated_initials)
                            if estimated_initials.count(name) > 1
                        ]
                    ]
                )
            )
        )

    search_param = np.empty(len(estimated_params) + len(estimated_initials))
    for i, j in enumerate(estimated_params):
        search_param[i] = param_values[j]
    for i, j in enumerate(estimated_initials):
        search_param[i + len(estimated_params)] = initial_values[j]

    if np.any(search_param == 0.0):
        message = "search_param must not contain zero."
        for idx in estimated_params:
            if param_values[int(idx)] == 0.0:
                raise ValueError(f"'C.{parameters[int(idx)]}' in self.idx_params: " + message)
        for idx in estimated_initials:
            if initial_values[int(idx)] == 0.0:
                raise ValueError(f"'V.{species[int(idx)]}' in self.idx_initials: " + message)
    return search_param


def _raise_value_error(
    parameters: List[str],
    species: List[str],
    search_param_index: int,
    message: str,
) -> NoReturn:
    """
    Raise ValueError when user-defined search_rgn is not appropriate.
    """
    if search_param_index <= len(parameters):
        raise ValueError(f"'C.{parameters[search_param_index]}': " + message)
    raise ValueError(f"'V.{species[search_param_index - len(parameters)]}': " + message)


def convert_scale(
    region: np.ndarray,
    parameters: List[str],
    species: List[str],
    estimated_params: List[int],
    estimated_initials: List[int],
) -> np.ndarray:
    """
    Convert search region from linear scale to logarithmic scale.

    Parameters
    ----------
    region : numpy ndarray
        Pre-converted search region.

    parameters : List[string]
        Names of model parameters.

    species : List[string]
        Names of model species.

    estimated_params : List[int]
        Indices of parameters to be estimated.

    estimated_initials : List[int]
        Indices of initial conditions to be estimated.

    Returns
    -------
    region : numpy ndarray
        Search bounds for parameters and/or initial conditions to be estimated.
    """
    for idx_search in range(region.shape[1]):
        if np.min(region[:, idx_search]) < 0.0:
            _raise_value_error(
                parameters,
                species,
                idx_search,
                "search_rgn[lower_bound, upper_bound] must be positive.",
            )
        elif np.min(region[:, idx_search]) == 0 and np.max(region[:, idx_search]) > 0:
            _raise_value_error(
                parameters,
                species,
                idx_search,
                "lower_bound must be larger than 0.",
            )
        elif region[0, idx_search] > region[1, idx_search]:
            _raise_value_error(
                parameters,
                species,
                idx_search,
                "lower_bound must be smaller than upper_bound.",
            )
    difference = list(
        set(np.where(np.any(region != 0.0, axis=0))[0])
        ^ set(np.append(estimated_params, [len(parameters) + idx for idx in estimated_initials]))
    )
    if len(difference) > 0:
        for idx_diff in difference:
            _raise_value_error(
                parameters,
                species,
                idx_diff,
                "should be set in both search_idx and search_rgn.",
            )

    search_rgn = region[:, np.any(region != 0.0, axis=0)]

    return np.log10(search_rgn)



import numpy as np
from scipy.spatial.distance import cosine

from .observable import ExperimentalData, NumericalSimulation, observables
from .set_search_param import SearchParam


def _compute_objval_rss(sim_data, exp_data):
    """Return Residual Sum of Squares"""
    return np.dot((sim_data - exp_data), (sim_data - exp_data))


def _compute_objval_cos(sim_data, exp_data):
    """Return Cosine distance"""
    return cosine(sim_data, exp_data)


def _diff_sim_and_exp(sim_matrix, exp_dict, exp_timepoint, conditions, sim_norm_max):
    sim_val = []
    exp_val = []

    for idx, condition in enumerate(conditions):
        if condition in exp_dict.keys():
            sim_val.extend(sim_matrix[list(map(int, exp_timepoint)), idx])
            exp_val.extend(exp_dict[condition])

    return np.array(sim_val) / sim_norm_max, np.array(exp_val)


def objective(indiv_gene, *args):
    """Define an objective function to be minimized"""
    if len(args) == 0:
        sp = SearchParam()
        indiv = sp.gene2val(indiv_gene)
        (x, y0) = sp.update(indiv)
    elif len(args) == 1:
        raise ValueError("not enough values to unpack (expected 2, got 1)")
    elif len(args) == 2:
        (x, y0) = args
    else:
        raise ValueError("too many values to unpack (expected 2)")

    sim = NumericalSimulation()
    exp = ExperimentalData()

    exp.set_data()

    if sim.simulate(x, y0) is None:
        error = np.zeros(len(observables))
        for i, obs_name in enumerate(observables):
            if exp.experiments[i] is not None:
                error[i] = _compute_objval_rss(
                    *_diff_sim_and_exp(
                        sim.simulations[i],
                        exp.experiments[i],
                        exp.get_timepoint(obs_name),
                        sim.conditions,
                        sim_norm_max=1
                        if not sim.normalization
                        else (
                            np.max(
                                sim.simulations[
                                    observables.index(obs_name),
                                    sim.normalization[obs_name]["timepoint"],
                                    [
                                        sim.conditions.index(c)
                                        for c in (
                                            sim.normalization[obs_name]["condition"]
                                            if sim.normalization[obs_name]["condition"]
                                            else sim.conditions
                                        )
                                    ],
                                ]
                            )
                            if sim.normalization[obs_name]["timepoint"] is not None
                            else np.max(
                                sim.simulations[
                                    observables.index(obs_name),
                                    :,
                                    [
                                        sim.conditions.index(c)
                                        for c in (
                                            sim.normalization[obs_name]["condition"]
                                            if sim.normalization[obs_name]["condition"]
                                            else sim.conditions
                                        )
                                    ],
                                ]
                            )
                        ),
                    )
                )
        return np.sum(error)  # < 1e12
    else:
        return 1e12



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "Ins",
    "pro_IRcom",
    "IRcom",
    "p1IRcom",
    "p2IRcom",
    "p1p2IRcom",
    "iAKT",
    "pAKT",
    "imTOR",
    "pmTOR",
    "iX",
    "pX",
    "iS6K",
    "pS6K",
    "iGSK3B",
    "pGSK3B",
    "iFoxO1",
    "pFoxO1",
    "G6Pase",
]

for idx, name in enumerate(NAMES):
    exec("{} = {:d}".format(name, idx))

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation

observables = [
    "pAKT",
    "pS6K",
    "pGSK3B",
    "G6Pase",
]


class NumericalSimulation(DifferentialEquation):
    """Simulate a model using scipy.integrate.ode
    Attributes
    ----------
    normalization : nested dict
        Keys for each observable
        ------------------------
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    """

    def __init__(self):
        super().__init__(perturbation={})
        self.normalization = {}

    t = range(480 + 1)  # min

    # Experimental conditions
    conditions = ["Insulin 0.01", "Insulin 0.03", "Insulin 0.1", "Insulin 0.3", "Insulin 1.0"]

    simulations = np.empty((len(observables), len(t), len(conditions)))

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation

        y0[V.Ins] = 0.01  # 0.01 nM of insulin during starvation
        sol = solve_ode(self.diffeq, y0, range(2400 + 1), tuple(x))
        if sol is None:
            return False
        else:
            y0 = sol.y[:, -1].tolist()

        for i, condition in enumerate(self.conditions):
            if condition == "Insulin 0.01":
                y0[V.Ins] = 0.01
            elif condition == "Insulin 0.03":
                y0[V.Ins] = 0.03
            elif condition == "Insulin 0.1":
                y0[V.Ins] = 0.1
            elif condition == "Insulin 0.3":
                y0[V.Ins] = 0.3
            elif condition == "Insulin 1.0":
                y0[V.Ins] = 1.0

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[observables.index("pAKT"), :, i] = sol.y[V.pAKT, :]
                self.simulations[observables.index("pS6K"), :, i] = (
                    sol.y[V.pS6K, :] * 83.8672192461257
                )
                self.simulations[observables.index("pGSK3B"), :, i] = (
                    sol.y[V.pGSK3B, :] * 0.111097316860158
                )
                self.simulations[observables.index("G6Pase"), :, i] = (
                    sol.y[V.G6Pase, :] * 0.0363622452066626
                )


class ExperimentalData(object):
    """
    Set experimental data.
    Attributes
    ----------
    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        self.experiments = [None] * len(observables)
        self.error_bars = [None] * len(observables)

    def set_data(self):
        pass

    @staticmethod
    def get_timepoint(obs_name):
        if obs_name in observables:
            return []



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation

observables = [
    "Per_mRNA",
    "Cry_mRNA",
    "Bmal1_mRNA",
]


class NumericalSimulation(DifferentialEquation):
    """Simulate a model using scipy.integrate.solve_ivp

    Attributes
    ----------
    normalization : nested dict
        Keys for each observable
        ------------------------
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    """

    def __init__(self):
        super().__init__(perturbation={})
        self.normalization = {}

    t = range(72 + 1)

    # Experimental conditions
    conditions = ["DD"]

    simulations = np.empty((len(observables), len(t), len(conditions)))

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation
        for i, condition in enumerate(self.conditions):
            if condition == "DD":
                pass

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[observables.index("Per_mRNA"), :, i] = sol.y[V.MP, :]
                self.simulations[observables.index("Cry_mRNA"), :, i] = sol.y[V.MC, :]
                self.simulations[observables.index("Bmal1_mRNA"), :, i] = sol.y[V.MB, :]


class ExperimentalData(object):
    """
    Set experimental data.

    Attributes
    ----------
    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        self.experiments = [None] * len(observables)
        self.error_bars = [None] * len(observables)

    def set_data(self):
        pass

    @staticmethod
    def get_timepoint(obs_name):
        if obs_name in observables:
            return []



from .name2idx import C, V


class DifferentialEquation(object):
    def __init__(self, perturbation):
        self.perturbation = perturbation

    def diffeq(self, t, y, *x):

        v = {}

        v[1] = x[C.k1_synthesis] * (y[V.pro_IRcom] - y[V.IRcom])
        v[2] = x[C.k1_InsIRcom] * y[V.Ins] * y[V.IRcom] - x[C.k2_InsIRcom] * y[V.p1IRcom]
        v[3] = x[C.k1_p1IRcomDeg] * y[V.p1IRcom]
        v[4] = x[C.k1_p1IRcomPhos] * y[V.pmTOR] * y[V.p1IRcom]
        v[5] = x[C.k1_p1p2IRcomdePhos] * y[V.p1p2IRcom]
        v[6] = x[C.k1_IRcomPhos] * y[V.pmTOR] * y[V.IRcom]
        v[7] = x[C.k1_p2IRcomdePhos] * y[V.p2IRcom]
        v[8] = x[C.k1_p2IRcomDeg] * y[V.p2IRcom]
        v[9] = x[C.k1_Insp2IRcom] * y[V.Ins] * y[V.p2IRcom] - x[C.k2_Insp2IRcom] * y[V.p1p2IRcom]
        v[10] = x[C.k1_p1p2IRcomDeg] * y[V.p1p2IRcom]
        v[11] = x[C.k1_AKTPhos] * (y[V.iAKT] - y[V.pAKT]) * y[V.p1IRcom]
        v[12] = x[C.k1_pAKTdePhos] * y[V.pAKT]
        v[13] = x[C.k1_mTORPhos] * (y[V.imTOR] - y[V.pmTOR]) * y[V.pAKT]
        v[14] = x[C.k1_pmTORdePhos] * y[V.pmTOR]
        v[15] = x[C.k1_S6KPhos] * (y[V.iS6K] - y[V.pS6K]) * y[V.pmTOR]
        v[16] = x[C.k1_pS6KdePhos] * y[V.pS6K] * y[V.pX]
        v[17] = x[C.k1_XPhos] * (y[V.iX] - y[V.pX]) * y[V.pmTOR]
        v[18] = x[C.k1_pXdePhos] * y[V.pX]
        v[19] = x[C.k1_GSK3BPhos] * (y[V.iGSK3B] - y[V.pGSK3B]) * y[V.pAKT]
        v[20] = x[C.k1_pGSK3BdePhos] * y[V.pGSK3B]
        v[21] = x[C.k1_FoxO1Phos] * (y[V.iFoxO1] - y[V.pFoxO1]) * y[V.pAKT]
        v[22] = x[C.k1_pFoxO1dePhos] * y[V.pFoxO1]
        v[23] = x[C.k1_G6PaseSynthesis] * (y[V.iFoxO1] - y[V.pFoxO1])
        v[24] = x[C.k1_G6PaseDeg] * y[V.G6Pase]

        """
        if self.perturbation:
            for i, dv in self.perturbation.items():
                v[i] = v[i] * dv
        """

        dydt = [0] * V.NUM

        dydt[V.IRcom] = v[1] - v[2] - v[6] + v[7]
        dydt[V.p1IRcom] = v[2] - v[3] - v[4] + v[5]
        dydt[V.p2IRcom] = v[6] - v[7] - v[8] - v[9]
        dydt[V.p1p2IRcom] = v[4] - v[5] + v[9] - v[10]
        dydt[V.pAKT] = v[11] - v[12]
        dydt[V.pmTOR] = v[13] - v[14]
        dydt[V.pX] = v[17] - v[18]
        dydt[V.pS6K] = v[15] - v[16]
        dydt[V.pGSK3B] = v[19] - v[20]
        dydt[V.pFoxO1] = v[21] - v[22]
        dydt[V.G6Pase] = v[23] - v[24]

        return dydt


def param_values():
    x = [0] * C.NUM

    x[C.k1_synthesis] = 0.04780
    x[C.k1_InsIRcom] = 7.78161
    x[C.k2_InsIRcom] = 1.61148
    x[C.k1_p1IRcomDeg] = 0.00792
    x[C.k1_p1IRcomPhos] = 0.00004
    x[C.k1_p1p2IRcomdePhos] = 0.28443
    x[C.k1_IRcomPhos] = 9.93311
    x[C.k1_p2IRcomdePhos] = 0.00001
    x[C.k1_p2IRcomDeg] = 0.00001
    x[C.k1_Insp2IRcom] = 0.36303
    x[C.k2_Insp2IRcom] = 0.40813
    x[C.k1_p1p2IRcomDeg] = 0.09490
    x[C.k1_AKTPhos] = 0.00920
    x[C.k1_pAKTdePhos] = 7.70619
    x[C.k1_mTORPhos] = 0.41968
    x[C.k1_pmTORdePhos] = 0.12433
    x[C.k1_S6KPhos] = 0.00752
    x[C.k1_pS6KdePhos] = 1.95498
    x[C.k1_XPhos] = 0.00105
    x[C.k1_pXdePhos] = 0.00146
    x[C.k1_GSK3BPhos] = 2.97538
    x[C.k1_pGSK3BdePhos] = 0.92460
    x[C.k1_FoxO1Phos] = 4.59698
    x[C.k1_pFoxO1dePhos] = 0.15172
    x[C.k1_G6PaseSynthesis] = 4.86146
    x[C.k1_G6PaseDeg] = 0.05496

    return x


def initial_values():
    y0 = [0] * V.NUM

    y0[V.pro_IRcom] = 46.22225
    y0[V.IRcom] = 46.22225
    y0[V.p1IRcom] = 0
    y0[V.p2IRcom] = 0
    y0[V.p1p2IRcom] = 0
    y0[V.iAKT] = 4.33812
    y0[V.pAKT] = 0
    y0[V.imTOR] = 0.09592
    y0[V.pmTOR] = 0
    y0[V.iX] = 14.99133
    y0[V.pX] = 0
    y0[V.iS6K] = 2.77699
    y0[V.pS6K] = 0
    y0[V.iGSK3B] = 10.56415
    y0[V.pGSK3B] = 0
    y0[V.iFoxO1] = 0.43571
    y0[V.pFoxO1] = 0
    y0[V.G6Pase] = 38.54029

    return y0



import os
import sys
from dataclasses import dataclass
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..exec_model import ExecModel, ModelObject
from .util import dlnyi_dlnxj, get_signaling_metric


@dataclass
class ParameterSensitivity(ExecModel):
    """Sensitivity for parameters"""

    model: ModelObject

    def _get_param_indices(self, excluded_params: List[str]) -> List[int]:
        param_indices = []
        x = self.model.pval()
        for i, val in enumerate(x):
            if self.model.parameters[i] not in excluded_params and val != 0.0:
                param_indices.append(i)
        if not param_indices:
            raise ValueError("No nonzero parameters.")

        return param_indices

    def _calc_sensitivity_coefficients(
        self,
        metric: str,
        param_indices: List[int],
        options,
    ) -> np.ndarray:
        """Calculating Sensitivity Coefficients

        Parameters
        ----------
        metric : str
            The signaling metric used for sensitivity analysis.

        param_indices : list of int
            List of parameter indices for sensitivity analysis.

        Returns
        -------
        sensitivity_coefficients : numpy array

        """

        rate = 1.01  # 1% change
        n_file = self.get_executable()

        signaling_metric = np.full(
            (
                len(n_file),
                len(param_indices) + 1,
                len(self.model.obs),
                len(self.model.sim.conditions),
            ),
            np.nan,
        )
        for i, nth_paramset in enumerate(n_file):
            optimized = self.load_param(nth_paramset)
            for j, idx in enumerate(param_indices):
                x = optimized.params[:]
                x[idx] = optimized.params[idx] * rate
                if self.model.sim.simulate(x, optimized.initials) is None:
                    for k, _ in enumerate(self.model.obs):
                        for l, _ in enumerate(self.model.sim.conditions):
                            signaling_metric[i, j, k, l] = get_signaling_metric(
                                metric, self.model.sim.simulations[k, :, l], options
                            )
                sys.stdout.write(
                    "\r{:d} / {:d}".format(
                        i * len(param_indices) + j + 1, len(n_file) * len(param_indices)
                    )
                )
            # Signaling metric without perturbation (j=-1)
            x = optimized.params[:]
            if self.model.sim.simulate(x, optimized.initials) is None:
                for k, _ in enumerate(self.model.obs):
                    for l, _ in enumerate(self.model.sim.conditions):
                        signaling_metric[i, -1, k, l] = get_signaling_metric(
                            metric, self.model.sim.simulations[k, :, l], options
                        )
        sensitivity_coefficients = dlnyi_dlnxj(
            signaling_metric,
            n_file,
            param_indices,
            self.model.obs,
            self.model.sim.conditions,
            rate,
        )

        return sensitivity_coefficients

    def _load_sc(
        self,
        metric: str,
        param_indices: List[int],
        options: dict,
    ) -> np.ndarray:
        """
        Load (or calculate) sensitivity coefficients.
        """
        if not os.path.isfile(
            os.path.join(
                self.model.path,
                "sensitivity_coefficients",
                "parameter",
                f"{metric}.npy",
            )
        ):
            os.makedirs(
                os.path.join(
                    self.model.path,
                    "sensitivity_coefficients",
                    "parameter",
                ),
                exist_ok=True,
            )
            sensitivity_coefficients = self._calc_sensitivity_coefficients(
                metric,
                param_indices,
                options,
            )
            np.save(
                os.path.join(
                    self.model.path,
                    "sensitivity_coefficients",
                    "parameter",
                    f"{metric}",
                ),
                sensitivity_coefficients,
            )
        else:
            sensitivity_coefficients = np.load(
                os.path.join(
                    self.model.path,
                    "sensitivity_coefficients",
                    "parameter",
                    f"{metric}.npy",
                )
            )
            if len(param_indices) != sensitivity_coefficients.shape[1]:
                # User changed options['excluded_params'] after the last trial
                sensitivity_coefficients = self._calc_sensitivity_coefficients(
                    metric,
                    param_indices,
                    options,
                )
                np.save(
                    os.path.join(
                        self.model.path,
                        "sensitivity_coefficients",
                        "parameter",
                        f"{metric}",
                    ),
                    sensitivity_coefficients,
                )

        return sensitivity_coefficients

    def _barplot_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        param_indices: List[int],
        save_format: str,
    ) -> None:
        """
        Visualize sensitivity coefficients using barplot.
        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "figure",
                "sensitivity",
                "parameter",
                f"{metric}",
                "barplot",
            ),
            exist_ok=True,
        )
        options = self.model.viz.sensitivity_options

        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        if len(options["cmap"]) < len(self.model.sim.conditions):
            raise ValueError(
                "len(sensitivity_options['cmap']) must be equal to "
                "or greater than len(sim.conditions)."
            )
        for k, obs_name in enumerate(self.model.obs):
            plt.figure(figsize=options["figsize"])
            plt.hlines([0], -options["width"], len(param_indices), "k", lw=1)
            for l, condition in enumerate(self.model.sim.conditions):
                sensitivity_matrix = sensitivity_coefficients[:, :, k, l]
                nan_idx = []
                for m in range(sensitivity_matrix.shape[0]):
                    if np.isnan(sensitivity_matrix[m, :]).any():
                        nan_idx.append(m)
                sensitivity_matrix = np.delete(sensitivity_matrix, nan_idx, axis=0)
                if sensitivity_matrix.size != 0:
                    average = np.mean(sensitivity_matrix, axis=0)
                    if sensitivity_matrix.shape[0] == 1:
                        stdev = np.zeros(sensitivity_matrix.shape[1])
                    else:
                        stdev = np.std(sensitivity_matrix, axis=0, ddof=1)
                    plt.bar(
                        np.arange(len(param_indices)) + l * options["width"],
                        average,
                        yerr=stdev,
                        ecolor=options["cmap"][l],
                        capsize=2,
                        width=options["width"],
                        color=options["cmap"][l],
                        align="center",
                        label=condition,
                    )
            plt.xticks(
                np.arange(len(param_indices))
                + options["width"] * 0.5 * (len(self.model.sim.conditions) - 1),
                [self.model.parameters[i] for i in param_indices],
                fontsize=6,
                rotation=90,
            )
            plt.ylabel(
                "Control coefficients on\n" + metric + " (" + obs_name.replace("_", " ") + ")"
            )
            plt.xlim(-options["width"], len(param_indices))
            plt.legend(loc=options["legend_loc"], frameon=False)
            plt.savefig(
                os.path.join(
                    self.model.path,
                    "figure",
                    "sensitivity",
                    "parameter",
                    f"{metric}",
                    "barplot",
                    f"{obs_name}.{save_format}",
                ),
                dpi=600 if save_format == "png" else None,
                bbox_inches="tight",
            )
            plt.close()

    @staticmethod
    def _remove_nan(sensitivity_matrix: np.ndarray, normalize: bool) -> np.ndarray:
        """
        Remove NaN from sensitivity matrix.
        """
        nan_idx = []
        for i in range(sensitivity_matrix.shape[0]):
            if np.isnan(sensitivity_matrix[i, :]).any():
                nan_idx.append(i)
            else:
                pass
            if np.nanmax(np.abs(sensitivity_matrix[i, :])) == 0.0:
                sensitivity_matrix[i, :] = np.zeros(sensitivity_matrix.shape[1])
            else:
                sensitivity_matrix[i, :] = sensitivity_matrix[i, :] / (
                    np.nanmax(np.abs(sensitivity_matrix[i, :])) if normalize else 1
                )

        return np.delete(sensitivity_matrix, nan_idx, axis=0)

    def _heatmap_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        param_indices: List[int],
        save_format: str,
    ) -> None:
        """
        Visualize sensitivity coefficients using heatmap.
        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "figure",
                "sensitivity",
                "parameter",
                f"{metric}",
                "heatmap",
            ),
            exist_ok=True,
        )
        options = self.model.viz.sensitivity_options
        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        for k, obs_name in enumerate(self.model.obs):
            for l, condition in enumerate(self.model.sim.conditions):
                sensitivity_matrix = self._remove_nan(
                    sensitivity_coefficients[:, :, k, l], normalize=False
                )
                if sensitivity_matrix.shape[0] > 1 and not np.all(sensitivity_matrix == 0.0):
                    g = sns.clustermap(
                        data=sensitivity_matrix,
                        center=0,
                        robust=True,
                        method="ward",
                        cmap="RdBu_r",
                        linewidth=0.5,
                        col_cluster=False,
                        figsize=options["figsize"],
                        xticklabels=[self.model.parameters[i] for i in param_indices],
                        yticklabels=[],
                        # cbar_kws={"ticks": [-1, 0, 1]}
                    )
                    cbar = g.ax_heatmap.collections[0].colorbar
                    cbar.ax.tick_params(labelsize=8)
                    plt.setp(g.ax_heatmap.get_xticklabels(), rotation=90)
                    plt.savefig(
                        os.path.join(
                            self.model.path,
                            "figure",
                            "sensitivity",
                            "parameter",
                            f"{metric}",
                            "heatmap",
                            f"{condition}_{obs_name}.{save_format}",
                        ),
                        dpi=600 if save_format == "png" else None,
                        bbox_inches="tight",
                    )
                    plt.close()

    def analyze(self, *, metric: str, style: str, save_format: str, options: dict) -> None:
        """
        Perform sensitivity analysis.
        """
        param_indices = self._get_param_indices(options["excluded_params"])
        sensitivity_coefficients = self._load_sc(metric, param_indices, options)
        if style == "barplot":
            self._barplot_sensitivity(
                metric,
                sensitivity_coefficients,
                param_indices,
                save_format,
            )
        elif style == "heatmap":
            if len(param_indices) < 2:
                pass
            else:
                self._heatmap_sensitivity(
                    metric,
                    sensitivity_coefficients,
                    param_indices,
                    save_format,
                )
        else:
            raise ValueError("Available styles are: 'barplot', 'heatmap'")



from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from .observable import observables


class Visualization(object):
    """
    Plotting parameters for customizing figure properties

    Attributes
    ----------
    cm : matplotlib.colors.ListedColormap (default: plt.cm.get_cmap('tab10'))
        Choosing colormaps for 'cmap'.

    timecourse_options : list of dict
        Plotting options for figure/simulation/<viz_type>/<each_observable>.

            Keys
            ----
            * 'divided_by' : int or float (default: 1)
                Convert time unit. (e.g. sec -> min).
            
            * 'xlim' : tuple
                Set the x limits of the current axes.
            
            * 'xticks' : list (default: None)
                Set the current tick locations of the x-axis.
            
            * 'xlabel' : str (default: 'Time')
                Set the label for the x-axis.
            
            * 'ylim' : tuple
                Set the y limits of the current axes.
            
            * 'yticks' : list (default: None)
                Set the current tick locations of the y-axis.
            
            * 'ylabel' : str (default: observables[i].replace('__', '\n').replace('_', ' '))
                Set the label for the y-axis.
            
            * 'exp_data' : bool (default: True)
                if False, experimental data will not be shown.
            
            * 'legend_loc' : Location String (default: None)
                Set the location of the legend. If 'legend_loc' is None,
                pyplot.legend will not be shown.
            
            * 'cmap' : list or tuple
                Set colormap.
            
            * 'shape' : list or tuple of strings (default: Line2D.filled_markers)
                Set markers.
            
            * 'dont_show' : list of strings
                Set conditions you don't want to plot.
            
    multiplot_options : dict
        Plotting options for figure/simulation/<viz_type>/multiplot_observables.

    sensitivity_options : dict
        Plotting options for figure/sensitivity.

    """
    def __init__(self):
        self.cm = plt.cm.get_cmap('tab10')
        
        self.timecourse_options = [
            {
                'divided_by' : 1,
                'xlim' : (),
                'xticks' : None,
                'xlabel': 'Time',
                'ylim' : (),
                'yticks' : None,
                'ylabel': observables[i].replace('__', '\n').replace('_', ' '),
                'exp_data' : True, 
                'legend_loc' : None,
                'cmap' : [self.cm.colors[j] for j in range(10)],
                'shape' : Line2D.filled_markers,
                'dont_show' : [],
            } for i, _ in enumerate(observables)]
        
        self.multiplot_options = {
            'fig_name' : 'multiplot_observables',
            'observables' : [],
            'condition' : None,
            'xlim' : (),
            'xticks' : None,
            'xlabel': 'Time',
            'ylim' : (),
            'yticks' : None,
            'ylabel': '',
            'cmap' : [self.cm.colors[j] for j in range(10)],
            'shape' : Line2D.filled_markers,
        }
    
        self.sensitivity_options = {
            'figsize' : (12, 5),
            'width' : 0.3,
            'legend_loc' : 'upper left',
            'cmap' : [self.cm.colors[j] for j in range(10)],
        }

    def get_timecourse_options(self):
        
        for i, _ in enumerate(observables):
            self.timecourse_options[i]['xticks'] = [200*i for i in range(4)]
            self.timecourse_options[i]['xlabel'] = 'time (min)'
        
        return self.timecourse_options

    def multiplot_observables(self):
        
        return self.multiplot_options
    
    @staticmethod
    def set_timecourse_rcParams():
        """ figure/simulation
        """
        plt.rcParams['font.size'] = 12
        plt.rcParams['axes.linewidth'] = 1.5
        plt.rcParams['xtick.major.width'] = 1.5
        plt.rcParams['ytick.major.width'] = 1.5
        plt.rcParams['lines.linewidth'] = 1.8
        plt.rcParams['lines.markersize'] = 12
        # plt.rcParams['font.family'] = 'Arial'
        # plt.rcParams['mathtext.fontset'] = 'custom'
        # plt.rcParams['mathtext.it'] = 'Arial:italic'

    @staticmethod
    def set_param_range_rcParams():
        """ figure/param_range
        """
        plt.rcParams['font.size'] = 12
        plt.rcParams['axes.linewidth'] = 1.2
        plt.rcParams['xtick.major.width'] = 1.2
        plt.rcParams['ytick.major.width'] = 1.2
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def set_sensitivity_rcParams():
        """ figure/sensitivity
        """
        plt.rcParams['font.size'] = 12
        plt.rcParams['axes.linewidth'] = 1.2
        plt.rcParams['xtick.major.width'] = 1.2
        plt.rcParams['ytick.major.width'] = 1.2
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def convert_species_name(name):
        """ figure/sensitivity/initial_condition
        - Sensitivity for species with nonzero initial conditions
        """
        return name


import os
import re
from dataclasses import dataclass
from typing import Any, List, NamedTuple

import numpy as np


class ModelObject(object):
    """
    The BioMASS model object.

    Parameters
    ----------
    path : str
        Path to a biomass model.
    biomass_model : Any
        A package containing biomass model properties.

    Attributes
    ----------
    path : str
        Path to the model.
    parameters : list of strings
        Names of model parameters.
    species : list of strings
        Names of model species.
    obs : list of strings
        Names of model observables.
    pval : Callable
        Numerical values of the parameters.
    ival : Callable
        Initial values.
    obj_func : Callable
        An objective function to be minimized for parameter estimation.
    sim : NumericalSimulation
        Simulation conditions and results.
    exp : ExperimentalData
        Experimental measurements used to estimate model parameters.
    viz : Visualization
        Plotting parameters for customizing figure properties.
    sp : SearchParam
        Model parameters to optimize and search region.
    rxn : ReactionNetwork
        Reaction indices grouped according to biological processes.
    """

    def __init__(
        self,
        path: str,
        biomass_model: Any,
    ):
        self._path = path
        self._parameters = biomass_model.C.NAMES
        self._species = biomass_model.V.NAMES
        self._obs = biomass_model.observables
        self.pval = biomass_model.param_values
        self.ival = biomass_model.initial_values
        self.obj_func = biomass_model.objective
        self.sim = biomass_model.NumericalSimulation()
        self.exp = biomass_model.ExperimentalData()
        self.viz = biomass_model.Visualization()
        self.sp = biomass_model.SearchParam()
        self.rxn = biomass_model.ReactionNetwork()

    @property
    def path(self) -> str:
        return self._path

    @property
    def parameters(self) -> List[str]:
        return self._parameters

    @property
    def species(self) -> List[str]:
        return self._species

    @property
    def obs(self) -> List[str]:
        duplicate = [name for name in set(self._obs) if self._obs.count(name) > 1]
        if not duplicate:
            return self._obs
        else:
            raise NameError(f"Duplicate observables: {', '.join(duplicate)}")


class OptimizedValues(NamedTuple):
    params: list
    initials: list


@dataclass
class ExecModel(object):
    model: ModelObject

    def get_individual(self, paramset: int) -> np.ndarray:
        best_generation = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{paramset:d}",
                "generation.npy",
            )
        )
        best_individual = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{paramset:d}",
                f"fit_param{int(best_generation):d}.npy",
            )
        )
        return best_individual

    def load_param(self, paramset: int) -> OptimizedValues:
        best_individual = self.get_individual(paramset)
        (x, y0) = self.model.sp.update(best_individual)
        return OptimizedValues(x, y0)

    def get_executable(self) -> List[int]:
        n_file = []
        try:
            fitparam_files = os.listdir(
                os.path.join(
                    self.model.path,
                    "out",
                )
            )
            for file in fitparam_files:
                if re.match(r"\d", file):
                    n_file.append(int(file))
            empty_folder = []
            for i, nth_paramset in enumerate(n_file):
                if not os.path.isfile(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        "generation.npy",
                    )
                ):
                    empty_folder.append(i)
            for i in sorted(empty_folder, reverse=True):
                n_file.pop(i)
        except FileNotFoundError as e:
            print(e)
        return n_file



import os
import sys
import warnings
from dataclasses import dataclass
from typing import List

import numpy as np

from ..exec_model import ModelObject
from .temporal_dynamics import TemporalDynamics


@dataclass
class SignalingSystems(TemporalDynamics):
    model: ModelObject

    def simulate_all(
        self,
        *,
        viz_type: str,
        show_all: bool,
        stdev: bool,
        save_format: str,
        param_range: dict,
    ) -> None:
        """
        Run simulation and save figures.
        """
        n_file: List[int] = [] if viz_type in ["original", "experiment"] else self.get_executable()
        simulations_all = np.full(
            (
                len(self.model.obs),
                len(n_file),
                len(self.model.sim.t),
                len(self.model.sim.conditions),
            ),
            np.nan,
        )
        if viz_type != "experiment":
            os.makedirs(
                os.path.join(
                    self.model.path,
                    "simulation_data",
                ),
                exist_ok=True,
            )
            if len(n_file) > 0:
                if len(n_file) == 1 and viz_type == "average":
                    raise ValueError(f"viz_type should be 'best', not '{viz_type}'.")
                for j, nth_paramset in enumerate(n_file):
                    if self._validate(nth_paramset):
                        for i, _ in enumerate(self.model.obs):
                            simulations_all[i, j, :, :] = self.model.sim.simulations[i, :, :]
                # simulations_all : numpy array
                # All simulated values with estimated parameter sets.
                self._save_simulations(viz_type, simulations_all)
                best_fitness_all = self._get_best_objval(n_file)
                best_paramset = n_file[np.argmin(best_fitness_all)]
                self._write_best_fit_param(best_paramset)
                if viz_type == "average":
                    pass
                elif viz_type == "best":
                    is_successful = self._validate(int(best_paramset))
                    if is_successful:
                        self._save_simulations(viz_type, self.model.sim.simulations)
                else:
                    # viz_type == 'n(=1,2,...)'
                    is_successful = self._validate(int(viz_type))
                    if is_successful:
                        self._save_simulations(viz_type, self.model.sim.simulations)
                # Visualization of estimated parameter values
                if 2 <= len(n_file):
                    popt = np.empty(
                        (
                            len(n_file),
                            len(self.model.sp.idx_params) + len(self.model.sp.idx_initials),
                        )
                    )
                    for i, nth_paramset in enumerate(n_file):
                        popt[i, :] = self.get_individual(nth_paramset)
                    self.plot_param_range(popt, save_format, **param_range)
            else:
                # viz_type == 'original'
                x = self.model.pval()
                y0 = self.model.ival()
                if self.model.sim.simulate(x, y0) is not None:
                    warnings.warn("Simulation failed. #original", RuntimeWarning)
                # simulations_original : numpy array
                # Simulated values with original parameter values.
                self._save_simulations(viz_type, self.model.sim.simulations)

        self.plot_timecourse(n_file, viz_type, show_all, stdev, save_format, simulations_all)

    def _preprocessing(self, simulated_values: np.ndarray) -> np.ndarray:
        """
        Replace small value in time-course simulated values to zero
        when all(abs(time_course) < sys.float_info.epsilon).

        Parameters
        ----------
        simulated_values : matrix (len(t) × len(self.model.sim.t))
        """
        if simulated_values.ndim == 2:
            for k, time_course in enumerate(simulated_values.T):
                if np.all(np.abs(time_course) < sys.float_info.epsilon):
                    simulated_values[:, k] = np.zeros(len(self.model.sim.t))
        elif simulated_values.ndim == 3:
            for i, _ in enumerate(self.model.obs):
                simulated_values[i] = self._preprocessing(simulated_values[i])
        elif simulated_values.ndim == 4:
            for j in range(simulated_values.shape[1]):
                for i, _ in enumerate(self.model.obs):
                    simulated_values[i, j] = self._preprocessing(simulated_values[i, j])
        return simulated_values

    def _save_simulations(self, viz_type: str, simulated_values: np.ndarray) -> None:
        """
        Save time course simulated values to simulation_data/.
        """
        np.save(
            os.path.join(
                self.model.path,
                "simulation_data",
                "simulations_{}.npy".format(viz_type if simulated_values.ndim < 4 else "all"),
            ),
            self._preprocessing(simulated_values),
        )

    def _get_best_objval(self, n_file: List[int]) -> np.ndarray:
        """
        Get best objective function values from parameter estimation results (out/).
        """
        best_fitness_all = np.full(len(n_file), np.inf)
        for i, nth_paramset in enumerate(n_file):
            if os.path.isfile(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "best_fitness.npy",
                )
            ):
                best_fitness_all[i] = np.load(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        "best_fitness.npy",
                    )
                )
        return best_fitness_all

    def _validate(self, nth_paramset: int) -> bool:
        """
        Validates the dynamical viability of a set of estimated parameter values.

        Parameters
        ----------
        nth_paramset : int
            Index of a parameter set.

        Returns
        -------
        is_successful : bool
            True if integration was successful.

        """
        optimized = self.load_param(nth_paramset)
        if self.model.sim.simulate(*optimized) is None:
            is_successful = True
        else:
            warnings.warn(f"Simulation failed. #{nth_paramset:d}", RuntimeWarning)
            is_successful = False

        return is_successful

    def _write_best_fit_param(
        self,
        best_paramset: int,
        indentation: str = " " * 4,
    ) -> None:
        """
        Create best_fit_param.txt in out/.

        Parameters
        ----------
        best_paramset : int
            Index of parameter set with the best objective value.

        """
        optimized = self.load_param(best_paramset)
        with open(
            os.path.join(
                self.model.path,
                "out",
                "best_fit_param.txt",
            ),
            mode="w",
        ) as f:
            f.write(f"# parameter set: {best_paramset:d}\n")
            f.write(f"\n\ndef param_values():\n{indentation}x = [0] * C.NUM\n")
            for i, name in enumerate(self.model.parameters):
                f.write(f"{indentation}x[C.{name}] = {optimized.params[i]:8.3e}\n")
            f.write(f"\n\ndef initial_values():\n{indentation}y0 = [0] * V.NUM\n")
            for i, name in enumerate(self.model.species):
                if optimized.initials[i] != 0:
                    f.write(f"{indentation}y0[V.{name}] = {optimized.initials[i]:8.3e}\n")



from typing import Dict, List

from biomass.analysis.reaction import is_duplicate


class ReactionNetwork(object):
    """
    Reaction indices grouped according to biological processes.
    This is used for sensitivity analysis (target='reaction').
    """

    def __init__(self) -> None:
        self.reactions: Dict[str, List[int]] = {
            "ERK_activation": [i for i in range(1, 7)],
            "ERK_dephosphorylation_by_DUSP": [i for i in range(47, 57)],
            "ERK_transport": [i for i in range(7, 10)],
            "RSK_activation": [24, 25],
            "RSK_transport": [26],
            "Elk1_activation": [29, 30],
            "CREB_activation": [27, 28],
            "dusp_production_etc": [i for i in range(10, 14)],
            "DUSP_transport": [18, 19],
            "DUSP_stabilization": [14, 15, 20, 21],
            "DUSP_degradation": [16, 17, 22, 23],
            "cfos_production_etc": [i for i in range(31, 35)],
            "cFos_transport": [40, 41],
            "cFos_stabilization": [35, 36, 37, 42, 43, 44],
            "cFos_degradation": [38, 39, 45, 46],
            "Feedback_from_F": [i for i in range(57, 64)],
        }

    def group(self):
        """
        Group reactions according to biological processes
        """
        biological_processes = []
        for process, indices in self.reactions.items():
            if not isinstance(indices, list):
                raise TypeError("Use list for reaction indices in {}".format(process))
            biological_processes.append(indices)

        if not is_duplicate(self.reactions, biological_processes):
            return biological_processes



from .fitness import objective
from .name2idx import C, V
from .observable import ExperimentalData, NumericalSimulation, observables
from .reaction_network import ReactionNetwork
from .set_model import initial_values, param_values
from .set_search_param import SearchParam
from .viz import Visualization



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "ppMEKc",
    "CREBn",
    "pCREBn",
    "ERKc",
    "ERKn",
    "pERKc",
    "pERKn",
    "ppERKc",
    "ppERKn",
    "Elk1n",
    "pElk1n",
    "cFOSc",
    "cFOSn",
    "pcFOSc",
    "pcFOSn",
    "DUSPc",
    "DUSPn",
    "pDUSPc",
    "pDUSPn",
    "DUSPn_ERKn",
    "DUSPn_pERKn",
    "DUSPn_ppERKn",
    "pDUSPn_ERKn",
    "pDUSPn_pERKn",
    "pDUSPn_ppERKn",
    "RSKc",
    "pRSKc",
    "pRSKn",
    "PrecfosmRNAn",
    "PreduspmRNAn",
    "cfosmRNAc",
    "duspmRNAc",
    "Fc",
    "Fn",
    "FmRNAc",
    "PreFmRNAn",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



import numpy as np
from scipy.spatial.distance import cosine

from .observable import observables, ExperimentalData, NumericalSimulation
from .set_search_param import SearchParam


def _compute_objval_rss(sim_data, exp_data):
    """Return Residual Sum of Squares"""
    return np.dot((sim_data - exp_data), (sim_data - exp_data))


def _compute_objval_cos(sim_data, exp_data):
    """Return Cosine distance"""
    return cosine(sim_data, exp_data)


def _diff_sim_and_exp(sim_matrix, exp_dict, exp_timepoint, conditions, sim_norm_max):
    sim_val = []
    exp_val = []

    for idx, condition in enumerate(conditions):
        if condition in exp_dict.keys():
            sim_val.extend(sim_matrix[list(map(int, exp_timepoint)), idx])
            exp_val.extend(exp_dict[condition])

    return np.array(sim_val) / sim_norm_max, np.array(exp_val)


def objective(indiv_gene, *args):
    """Define an objective function to be minimized"""
    if len(args) == 0:
        sp = SearchParam()
        indiv = sp.gene2val(indiv_gene)
        (x, y0) = sp.update(indiv)
    elif len(args) == 1:
        raise ValueError("not enough values to unpack (expected 2, got 1)")
    elif len(args) == 2:
        (x, y0) = args
    else:
        raise ValueError("too many values to unpack (expected 2)")

    sim = NumericalSimulation()
    exp = ExperimentalData()

    exp.set_data()

    if sim.simulate(x, y0) is None:
        error = np.zeros(len(observables))
        for i, obs_name in enumerate(observables):
            if exp.experiments[i] is not None:
                error[i] = _compute_objval_rss(
                    *_diff_sim_and_exp(
                        sim.simulations[i],
                        exp.experiments[i],
                        exp.get_timepoint(obs_name),
                        sim.conditions,
                        sim_norm_max=1
                        if not sim.normalization
                        else (
                            np.max(
                                sim.simulations[
                                    observables.index(obs_name),
                                    sim.normalization[obs_name]["timepoint"],
                                    [
                                        sim.conditions.index(c)
                                        for c in (
                                            sim.normalization[obs_name]["condition"]
                                            if sim.normalization[obs_name]["condition"]
                                            else sim.conditions
                                        )
                                    ],
                                ]
                            )
                            if sim.normalization[obs_name]["timepoint"] is not None
                            else np.max(
                                sim.simulations[
                                    observables.index(obs_name),
                                    :,
                                    [
                                        sim.conditions.index(c)
                                        for c in (
                                            sim.normalization[obs_name]["condition"]
                                            if sim.normalization[obs_name]["condition"]
                                            else sim.conditions
                                        )
                                    ],
                                ]
                            )
                        ),
                    )
                )
        """
        error = np.zeros(16)

        norm_max = np.max(sim.simulations[observables.index('Phosphorylated_MEKc')])
        error[0] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_MEKc'), exp.t2, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_MEKc')]['EGF']
        )
        error[1] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_MEKc'), exp.t2, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_MEKc')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('Phosphorylated_ERKc')])
        error[2] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_ERKc'), exp.t2, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_ERKc')]['EGF']
        )
        error[3] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_ERKc'), exp.t2, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_ERKc')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('Phosphorylated_RSKw')])
        error[4] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_RSKw'), exp.t2, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_RSKw')]['EGF']
        )
        error[5] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_RSKw'), exp.t2, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_RSKw')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('Phosphorylated_CREBw')])
        error[6] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_CREBw'), exp.t3, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_CREBw')]['EGF']
        )
        error[7] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_CREBw'), exp.t3, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_CREBw')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('dusp_mRNA')])
        error[8] = _compute_objval_rss(
            sim.simulations[observables.index('dusp_mRNA'), exp.t5, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('dusp_mRNA')]['EGF']
        )
        error[9] = _compute_objval_rss(
            sim.simulations[observables.index('dusp_mRNA'), exp.t5, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('dusp_mRNA')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('cfos_mRNA')])
        error[10] = _compute_objval_rss(
            sim.simulations[observables.index('cfos_mRNA'), exp.t4, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('cfos_mRNA')]['EGF']
        )
        error[11] = _compute_objval_rss(
            sim.simulations[observables.index('cfos_mRNA'), exp.t4, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('cfos_mRNA')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('cFos_Protein')])
        error[12] = _compute_objval_rss(
            sim.simulations[observables.index('cFos_Protein'), exp.t5, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('cFos_Protein')]['EGF']
        )
        error[13] = _compute_objval_rss(
            sim.simulations[observables.index('cFos_Protein'), exp.t5, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('cFos_Protein')]['HRG']
        )

        norm_max = np.max(sim.simulations[observables.index('Phosphorylated_cFos')])
        error[14] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_cFos'), exp.t2, sim.conditions.index('EGF')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_cFos')]['EGF']
        )
        error[15] = _compute_objval_rss(
            sim.simulations[observables.index('Phosphorylated_cFos'), exp.t2, sim.conditions.index('HRG')]/norm_max, 
            exp.experiments[observables.index('Phosphorylated_cFos')]['HRG']
        )
        """
        return np.sum(error)  # < 1e12
    else:
        return 1e12


from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from .observable import observables


class Visualization(object):
    """
    Plotting parameters for customizing figure properties

    Attributes
    ----------
    cm : matplotlib.colors.ListedColormap (default: plt.cm.get_cmap('tab10'))
        Choosing colormaps for 'cmap'.

    timecourse_options : list of dict
        Plotting options for figure/simulation/<viz_type>/<each_observable>.

            Keys
            ----
            * 'divided_by' : int or float (default: 1)
                Convert time unit. (e.g. sec -> min).

            * 'xlim' : tuple
                Set the x limits of the current axes.

            * 'xticks' : list (default: None)
                Set the current tick locations of the x-axis.

            * 'xlabel' : str (default: 'Time')
                Set the label for the x-axis.

            * 'ylim' : tuple
                Set the y limits of the current axes.

            * 'yticks' : list (default: None)
                Set the current tick locations of the y-axis.

            * 'ylabel' : str (default: observables[i].replace('__', '\n').replace('_', ' '))
                Set the label for the y-axis.

            * 'exp_data' : bool (default: True)
                if False, experimental data will not be shown.

            * 'legend_loc' : Location String (default: None)
                Set the location of the legend. If 'legend_loc' is None,
                pyplot.legend will not be shown.

            * 'cmap' : list or tuple
                Set colormap.

            * 'shape' : list or tuple of strings (default: Line2D.filled_markers)
                Set markers.

            * 'dont_show' : list of strings
                Set conditions you don't want to plot.

    multiplot_options : dict
        Plotting options for figure/simulation/<viz_type>/multiplot_observables.

    sensitivity_options : dict
        Plotting options for figure/sensitivity.

    """

    def __init__(self):
        self.cm = plt.cm.get_cmap("tab10")

        self.timecourse_options = [
            {
                "divided_by": 1,
                "xlim": (),
                "xticks": None,
                "xlabel": "Time",
                "ylim": (),
                "yticks": None,
                "ylabel": observables[i].replace("__", "\n").replace("_", " "),
                "exp_data": True,
                "legend_loc": None,
                "cmap": [self.cm.colors[j] for j in range(10)],
                "shape": Line2D.filled_markers,
                "dont_show": [],
            }
            for i, _ in enumerate(observables)
        ]

        self.multiplot_options = {
            "fig_name": "multiplot_observables",
            "observables": [],
            "condition": None,
            "xlim": (),
            "xticks": None,
            "xlabel": "Time",
            "ylim": (),
            "yticks": None,
            "ylabel": "",
            "cmap": [self.cm.colors[j] for j in range(10)],
            "shape": Line2D.filled_markers,
        }

        self.sensitivity_options = {
            "figsize": (12, 5),
            "width": 0.3,
            "legend_loc": "upper left",
            "cmap": [self.cm.colors[j] for j in range(10)],
        }

    def get_timecourse_options(self):

        for i, _ in enumerate(observables):
            self.timecourse_options[i]["xticks"] = [120 * i for i in range(5)]
            self.timecourse_options[i]["xlabel"] = "time (min)"
            self.timecourse_options[i]["yticks"] = [0.2 * i for i in range(6)]

        return self.timecourse_options

    def multiplot_observables(self):

        return self.multiplot_options

    @staticmethod
    def set_timecourse_rcParams():
        """figure/simulation"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["xtick.major.width"] = 1.5
        plt.rcParams["ytick.major.width"] = 1.5
        plt.rcParams["lines.linewidth"] = 1.8
        plt.rcParams["lines.markersize"] = 12
        # plt.rcParams['font.family'] = 'Arial'
        # plt.rcParams['mathtext.fontset'] = 'custom'
        # plt.rcParams['mathtext.it'] = 'Arial:italic'

    @staticmethod
    def set_param_range_rcParams():
        """figure/param_range"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.2
        plt.rcParams["xtick.major.width"] = 1.2
        plt.rcParams["ytick.major.width"] = 1.2
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def set_sensitivity_rcParams():
        """figure/sensitivity"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.2
        plt.rcParams["xtick.major.width"] = 1.2
        plt.rcParams["ytick.major.width"] = 1.2
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def convert_species_name(name):
        """figure/sensitivity/initial_condition
        - Sensitivity for species with nonzero initial conditions
        """
        return name



import os
import warnings
from dataclasses import dataclass
from math import isnan
from typing import List, Optional, Tuple

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes._axes import _log as matplotlib_axes_logger

from ..exec_model import ExecModel, ModelObject

matplotlib_axes_logger.setLevel("ERROR")


class VisualizeWarning(UserWarning):
    pass


def _check_unknown_options(unknown_options: dict) -> None:
    if unknown_options:
        msg = ", ".join(map(str, unknown_options.keys()))
        warnings.warn(f"Unknown visualization options: {msg}", VisualizeWarning)


@dataclass
class TemporalDynamics(ExecModel):
    model: ModelObject

    def plot_timecourse(
        self,
        n_file: List[int],
        viz_type: str,
        show_all: bool,
        stdev: bool,
        save_format: str,
        simulations_all: np.ndarray,
    ) -> None:
        """
        Plot time course of each observable.

        Parameters
        ----------
        n_file : list of integers
            Optimized parameter sets in out/.

        viz_type : str
            One of ['average', 'best', 'original', 'n(=1,2,...)', 'experiment'].

        show_all : bool
            Whether to show all simulation results.

        stdev : bool
            If True, the standard deviation of simulated values will be shown
            (only available for 'average' visualization type).

        save_format : str (default: "pdf")
            Either "png" or "pdf", indicating whether to save figures
            as png or pdf format.

        simulations_all : numpy array
            Array containing all simulated values.

        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
            ),
            exist_ok=True,
        )
        self.model.exp.set_data()
        self.model.viz.set_timecourse_rcParams()
        singleplot = self.model.viz.get_timecourse_options()
        multiplot = self.model.viz.multiplot_observables()

        for mode in range(2):
            # mode 0 : timecourse_for_each_observable
            # mode 1 : multiplot_observables
            set_fig = False
            for i, obs_name in enumerate(self.model.obs):
                if len(singleplot[i]["cmap"]) < len(self.model.sim.conditions) or len(
                    singleplot[i]["shape"]
                ) < len(self.model.sim.conditions):
                    raise ValueError(
                        "len(cmap), len(shape) must be equal to"
                        " or greater than len(self.model.sim.conditions)."
                    )
                if mode == 1 and obs_name not in multiplot["observables"]:
                    continue
                if mode == 0:
                    plt.figure(figsize=(4, 3))
                elif mode == 1 and not set_fig:
                    plt.figure(figsize=(4, 3))
                    set_fig = True
                plt.gca().spines["right"].set_visible(False)
                plt.gca().spines["top"].set_visible(False)
                if viz_type != "experiment":
                    if show_all:
                        self._plot_show_all(
                            n_file, simulations_all, obs_name, mode, singleplot, multiplot
                        )
                    if viz_type == "average":
                        normalized = self._normalize_array(
                            n_file, simulations_all, obs_name, mode, singleplot, multiplot
                        )
                        if (
                            self.model.sim.normalization
                            and self.model.sim.normalization[obs_name]["timepoint"] is None
                        ):
                            normalized = self._divide_by_maximum(normalized, obs_name)
                        self._plot_average(normalized, obs_name, mode, singleplot, multiplot)
                        if stdev:
                            self._show_sd(normalized, obs_name, mode, singleplot, multiplot)
                    else:
                        self._plot_simulations(obs_name, mode, singleplot, multiplot)
                if (
                    viz_type == "experiment" or singleplot[i]["exp_data"]
                ) and self.model.exp.experiments[i] is not None:
                    exp_t = self.model.exp.get_timepoint(obs_name)
                    if self.model.exp.error_bars[i] is not None:
                        self._plot_experimental_data_with_error_bars(
                            viz_type, exp_t, obs_name, mode, singleplot, multiplot
                        )
                    else:
                        self._plot_experimental_data_without_error_bars(
                            viz_type, exp_t, obs_name, mode, singleplot, multiplot
                        )
                if mode == 0:
                    self._save_mode_0(obs_name, singleplot, viz_type, save_format)
            if mode == 1 and multiplot["observables"]:
                self._save_mode_1(multiplot, viz_type, save_format)

    def plot_param_range(
        self,
        popt: np.ndarray,
        save_format: str,
        orientation: str,
        distribution: str,
        scatter: bool,
        **unknown_options,
    ) -> None:
        """
        Plot estimated parameter/initial values.

        Parameters
        ----------
        popt : numpy array
            Array containing all optimized parameter/initial values.

        save_format : str (default: "pdf")
            Either "png" or "pdf", indicating whether to save figures
            as png or pdf format.

        orientation : str (default: 'portrait')
            Either 'portrait' or 'landscape'.

        distribution : str (default: 'boxenplot')
            Either 'boxplot' or 'boxenplot'.

        scatter : bool (default: False)
            If True, draw a stripplot.

        """
        _check_unknown_options(unknown_options)
        os.makedirs(
            os.path.join(
                self.model.path,
                "figure",
                "param_range",
            ),
            exist_ok=True,
        )

        self.model.viz.set_param_range_rcParams()

        for val in ["parameter_value", "initial_value"]:
            if (val == "parameter_value" and len(self.model.sp.idx_params) == 0) or (
                val == "initial_value" and len(self.model.sp.idx_initials) == 0
            ):
                continue
            else:
                fig = plt.figure(figsize=self._set_figsize(orientation, val))
                arguments_distribution = {
                    "data": popt[:, : len(self.model.sp.idx_params)]
                    if val == "parameter_value"
                    else popt[:, len(self.model.sp.idx_params) :],
                    "orient": "h" if orientation == "portrait" else "v",
                    "linewidth": 0.5,
                    "palette": "Set2",
                }
                arguments_scatter = {
                    "data": popt[:, : len(self.model.sp.idx_params)]
                    if val == "parameter_value"
                    else popt[:, len(self.model.sp.idx_params) :],
                    "orient": "h" if orientation == "portrait" else "v",
                    "size": 2.5,
                    "color": ".26",
                }
                if distribution == "boxplot":
                    ax = sns.boxplot(**arguments_distribution)
                else:
                    ax = sns.boxenplot(**arguments_distribution)
                if scatter:
                    ax = sns.stripplot(**arguments_scatter)
                sns.despine()
                if orientation == "portrait":
                    ax.set_xlabel(val.capitalize().replace("_", " "))
                    ax.set_xscale("log")
                    ax.set_ylabel("")
                    ax.set_yticklabels(
                        [self.model.parameters[i] for i in self.model.sp.idx_params]
                        if val == "parameter_value"
                        else [self.model.species[i] for i in self.model.sp.idx_initials]
                    )
                else:
                    ax.set_xlabel("")
                    ax.set_xticklabels(
                        [self.model.parameters[i] for i in self.model.sp.idx_params]
                        if val == "parameter_value"
                        else [self.model.species[i] for i in self.model.sp.idx_initials],
                        rotation=90,
                    )
                    ax.set_ylabel(val.capitalize().replace("_", " "))
                    ax.set_yscale("log")
                plt.savefig(
                    os.path.join(
                        self.model.path,
                        "figure",
                        "param_range",
                        f"estimated_{val}s." + save_format,
                    ),
                    dpi=600 if save_format == "png" else None,
                    bbox_inches="tight",
                )
                plt.close(fig)

    def _plot_show_all(
        self,
        n_file: List[int],
        simulations_all: np.ndarray,
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot time course simulated values (show_all == True).
        """
        i = self.model.obs.index(obs_name)
        for j, _ in enumerate(n_file):
            for l, condition in enumerate(self.model.sim.conditions):
                if (mode == 0 and condition not in singleplot[i]["dont_show"]) or (
                    mode == 1 and condition == multiplot["condition"]
                ):
                    plt.plot(
                        np.array(self.model.sim.t) / singleplot[i]["divided_by"],
                        simulations_all[i, j, :, l]
                        / (
                            1
                            if not self.model.sim.normalization
                            or np.max(simulations_all[i, j, :, l]) == 0.0
                            else np.max(
                                simulations_all[
                                    i,
                                    j,
                                    self.model.sim.normalization[obs_name]["timepoint"],
                                    [
                                        self.model.sim.conditions.index(c)
                                        for c in self.model.sim.normalization[obs_name][
                                            "condition"
                                        ]
                                    ],
                                ]
                            )
                            if self.model.sim.normalization[obs_name]["timepoint"] is not None
                            else np.max(
                                simulations_all[
                                    i,
                                    j,
                                    :,
                                    [
                                        self.model.sim.conditions.index(c)
                                        for c in self.model.sim.normalization[obs_name][
                                            "condition"
                                        ]
                                    ],
                                ]
                            )
                        ),
                        color=singleplot[i]["cmap"][l]
                        if mode == 0
                        else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                        alpha=0.05,
                    )

    def _normalize_array(
        self,
        n_file: List[int],
        simulations_all: np.ndarray,
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> np.ndarray:
        """
        Normalize the array simulations_all using sim.normalization set in observable.py.
        """
        normalized: np.ndarray = np.empty_like(simulations_all)
        i = self.model.obs.index(obs_name)
        for j, _ in enumerate(n_file):
            for l, condition in enumerate(self.model.sim.conditions):
                if (mode == 0 and condition not in singleplot[i]["dont_show"]) or (
                    mode == 1 and condition == multiplot["condition"]
                ):
                    normalized[i, j, :, l] = simulations_all[i, j, :, l] / (
                        1
                        if not self.model.sim.normalization
                        or np.max(simulations_all[i, j, :, l]) == 0.0
                        else np.max(
                            simulations_all[
                                i,
                                j,
                                self.model.sim.normalization[obs_name]["timepoint"],
                                [
                                    self.model.sim.conditions.index(c)
                                    for c in self.model.sim.normalization[obs_name]["condition"]
                                ],
                            ]
                        )
                        if self.model.sim.normalization[obs_name]["timepoint"] is not None
                        else np.max(
                            simulations_all[
                                i,
                                j,
                                :,
                                [
                                    self.model.sim.conditions.index(c)
                                    for c in self.model.sim.normalization[obs_name]["condition"]
                                ],
                            ]
                        )
                    )
        return normalized

    def _divide_by_maximum(self, normalized: np.ndarray, obs_name: str) -> np.ndarray:
        """
        Divide the array by its maximum.
        """
        mean_vec = []
        for c in self.model.sim.normalization[obs_name]["condition"]:
            mean_vec.append(
                np.nanmean(
                    normalized[
                        self.model.obs.index(obs_name), :, :, self.model.sim.conditions.index(c)
                    ],
                    axis=0,
                )
            )
        norm_max = np.max(mean_vec)
        if not isnan(norm_max) and norm_max != 0.0:
            normalized[self.model.obs.index(obs_name), :, :, :] /= norm_max

        return normalized

    def _plot_average(
        self,
        normalized: np.ndarray,
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot time course simulated values (viz_type == 'average').
        """
        i = self.model.obs.index(obs_name)
        for l, condition in enumerate(self.model.sim.conditions):
            if (mode == 0 and condition not in singleplot[i]["dont_show"]) or (
                mode == 1 and condition == multiplot["condition"]
            ):
                plt.plot(
                    np.array(self.model.sim.t) / singleplot[i]["divided_by"],
                    np.nanmean(normalized[i, :, :, l], axis=0),
                    color=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    label=condition if mode == 0 else singleplot[i]["ylabel"],
                )

    def _show_sd(
        self,
        normalized: np.ndarray,
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot standard deviation (SD) as shaded area when stdev == True.
        """
        i = self.model.obs.index(obs_name)
        for l, condition in enumerate(self.model.sim.conditions):
            if (mode == 0 and condition not in singleplot[i]["dont_show"]) or (
                mode == 1 and condition == multiplot["condition"]
            ):
                y_mean = np.nanmean(normalized[i, :, :, l], axis=0)
                y_std = [
                    np.nanstd(normalized[i, :, k, l], ddof=1)
                    for k, _ in enumerate(self.model.sim.t)
                ]
                plt.fill_between(
                    np.array(self.model.sim.t) / singleplot[i]["divided_by"],
                    y_mean - y_std,
                    y_mean + y_std,
                    lw=0,
                    color=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    alpha=0.1,
                )

    def _plot_simulations(
        self,
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot time course simulated values (viz_type not in ['average', 'experiment']).
        """
        i = self.model.obs.index(obs_name)
        for l, condition in enumerate(self.model.sim.conditions):
            if (mode == 0 and condition not in singleplot[i]["dont_show"]) or (
                mode == 1 and condition == multiplot["condition"]
            ):
                plt.plot(
                    np.array(self.model.sim.t) / singleplot[i]["divided_by"],
                    self.model.sim.simulations[i, :, l]
                    / (
                        1
                        if not self.model.sim.normalization
                        or np.max(self.model.sim.simulations[i, :, l]) == 0.0
                        else np.max(
                            self.model.sim.simulations[
                                i,
                                self.model.sim.normalization[obs_name]["timepoint"],
                                [
                                    self.model.sim.conditions.index(c)
                                    for c in self.model.sim.normalization[obs_name]["condition"]
                                ],
                            ]
                        )
                        if self.model.sim.normalization[obs_name]["timepoint"] is not None
                        else np.max(
                            self.model.sim.simulations[
                                i,
                                :,
                                [
                                    self.model.sim.conditions.index(c)
                                    for c in self.model.sim.normalization[obs_name]["condition"]
                                ],
                            ]
                        )
                    ),
                    color=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    label=condition if mode == 0 else singleplot[i]["ylabel"],
                )

    def _plot_experimental_data_with_error_bars(
        self,
        viz_type: str,
        exp_t: Optional[List[int]],
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot experimental measurements with error bars.
        """
        i = self.model.obs.index(obs_name)
        for l, condition in enumerate(self.model.sim.conditions):
            if (
                condition in self.model.exp.experiments[i]
                and (mode == 0 and condition not in singleplot[i]["dont_show"])
                or (mode == 1 and condition == multiplot["condition"])
            ):
                exp_data = plt.errorbar(
                    np.array(exp_t) / singleplot[i]["divided_by"],
                    self.model.exp.experiments[i][condition],
                    yerr=self.model.exp.error_bars[i][condition],
                    color=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    ecolor=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    elinewidth=1,
                    capsize=8,
                    markerfacecolor="None",
                    markeredgecolor=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    fmt=singleplot[i]["shape"][l]
                    if mode == 0
                    else multiplot["shape"][multiplot["observables"].index(obs_name)],
                    clip_on=False,
                    label=singleplot[i]["ylabel"]
                    if mode == 1 and viz_type == "experiment"
                    else None,
                )
                for capline in exp_data[1]:
                    capline.set_clip_on(False)
                for barlinecol in exp_data[2]:
                    barlinecol.set_clip_on(False)

    def _plot_experimental_data_without_error_bars(
        self,
        viz_type: str,
        exp_t: Optional[List[int]],
        obs_name: str,
        mode: int,
        singleplot: List[dict],
        multiplot: dict,
    ) -> None:
        """
        Plot experimental measurements when model.exp.error_bars[i] is None.
        """
        i = self.model.obs.index(obs_name)
        for l, condition in enumerate(self.model.sim.conditions):
            if (
                condition in self.model.exp.experiments[i]
                and (mode == 0 and condition not in singleplot[i]["dont_show"])
                or (mode == 1 and condition == multiplot["condition"])
            ):
                plt.plot(
                    np.array(exp_t) / singleplot[i]["divided_by"],
                    self.model.exp.experiments[i][condition],
                    singleplot[i]["shape"][l]
                    if mode == 0
                    else multiplot["shape"][multiplot["observables"].index(obs_name)],
                    markerfacecolor="None",
                    markeredgecolor=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    color=singleplot[i]["cmap"][l]
                    if mode == 0
                    else multiplot["cmap"][multiplot["observables"].index(obs_name)],
                    clip_on=False,
                    label=singleplot[i]["ylabel"]
                    if mode == 1 and viz_type == "experiment"
                    else None,
                )

    def _save_mode_0(
        self,
        obs_name: str,
        singleplot: List[dict],
        viz_type: str,
        save_format: str,
    ) -> None:
        """
        Plot time course of each observable.
        """
        i = self.model.obs.index(obs_name)
        if singleplot[i]["xlim"]:
            plt.xlim(singleplot[i]["xlim"])
        if singleplot[i]["xticks"] is not None:
            plt.xticks(singleplot[i]["xticks"])
        plt.xlabel(singleplot[i]["xlabel"])
        if singleplot[i]["ylim"]:
            plt.ylim(singleplot[i]["ylim"])
        if singleplot[i]["yticks"] is not None:
            plt.yticks(singleplot[i]["yticks"])
        plt.ylabel(singleplot[i]["ylabel"])
        if singleplot[i]["legend_loc"] is not None:
            plt.legend(
                loc=singleplot[i]["legend_loc"],
                frameon=False,
                fontsize=12,
            )
        plt.savefig(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
                f"{obs_name}." + save_format,
            ),
            dpi=600 if save_format == "png" else None,
            bbox_inches="tight",
        )
        plt.close()

    def _save_mode_1(
        self,
        multiplot: dict,
        viz_type: str,
        save_format: str,
    ) -> None:
        """
        Plot time course of multiple observables in one figure.
        """
        if multiplot["xlim"]:
            plt.xlim(multiplot["xlim"])
        if multiplot["xticks"] is not None:
            plt.xticks(multiplot["xticks"])
        plt.xlabel(multiplot["xlabel"])
        if multiplot["ylim"]:
            plt.ylim(multiplot["ylim"])
        if multiplot["yticks"] is not None:
            plt.yticks(multiplot["yticks"])
        plt.ylabel(multiplot["ylabel"])
        plt.legend(
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
            borderaxespad=0,
            labelspacing=1.25,
            frameon=False,
            fontsize=12,
        )
        plt.savefig(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
                f"{multiplot['fig_name']}." + save_format,
            ),
            dpi=600 if save_format == "png" else None,
            bbox_inches="tight",
        )
        plt.close()

    def _set_figsize(self, orientation: str, val: str) -> Tuple[float, float]:
        figsize = (
            (
                8.0,
                (
                    len(self.model.sp.idx_params)
                    if val == "parameter_value"
                    else len(self.model.sp.idx_initials)
                )
                / 2.5,
            )
            if orientation == "portrait"
            else (
                (
                    len(self.model.sp.idx_params)
                    if val == "parameter_value"
                    else len(self.model.sp.idx_initials)
                )
                / 2.2,
                6.0,
            )
        )
        return figsize
